# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


def literal(value):
    return {"type": "literal", "children": [], "value": value}


def identifier(name):
    return {"type": "identifier", "children": [], "value": name}


def path_identifier(left, right):
    return {"type": "path_identifier", "children": [left, right]}


def index_identifier(name, index):
    return {"type": "index_identifier", "children": [name], "value": index}


def sequence(elements):
    return {"type": "sequence", "children": elements}


def or_expression(left, right):
    return {"type": "or_expression", "children": [left, right]}


def and_expression(left, right):
    return {"type": "and_expression", "children": [left, right]}


def not_expression(expression):
    return {"type": "not_expression", "children": [expression]}


def subexpression(expression):
    return {"type": "subexpression", "children": [expression]}


def function_expression(name, arguments):
    return {"type": "function", "children": arguments, "value": name}


def in_expression(left, right):
    return {"type": "in_expression", "children": [left, right]}


def between_expression(left, center, right):
    return {"type": "between_expression", "children": [left, center, right]}


def comparison_expression(name, left, right):
    return {"type": "comparator", "children": [left, right], "value": name}



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscli.compat import is_windows
from awscli.customizations.commands import BasicCommand
from awscli.customizations.history.constants import (
    DEFAULT_HISTORY_FILENAME,
    HISTORY_FILENAME_ENV_VAR,
)
from awscli.customizations.history.db import (
    DatabaseConnection,
    DatabaseRecordReader,
)
from awscli.utils import OutputStreamFactory, is_a_tty


class HistorySubcommand(BasicCommand):
    def __init__(self, session, db_reader=None, output_stream_factory=None):
        super(HistorySubcommand, self).__init__(session)
        self._db_reader = db_reader
        self._output_stream_factory = output_stream_factory
        if output_stream_factory is None:
            self._output_stream_factory = (
                self._get_default_output_stream_factory()
            )

    def _get_default_output_stream_factory(self):
        return OutputStreamFactory(self._session)

    def _connect_to_history_db(self):
        if self._db_reader is None:
            connection = DatabaseConnection(self._get_history_db_filename())
            self._db_reader = DatabaseRecordReader(connection)

    def _close_history_db(self):
        self._db_reader.close()

    def _get_history_db_filename(self):
        filename = os.environ.get(
            HISTORY_FILENAME_ENV_VAR, DEFAULT_HISTORY_FILENAME
        )
        if not os.path.exists(filename):
            raise RuntimeError(
                'Could not locate history. Make sure cli_history is set to '
                'enabled in the ~/.aws/config file'
            )
        return filename

    def _should_use_color(self, parsed_globals):
        if parsed_globals.color == 'on':
            return True
        elif parsed_globals.color == 'off':
            return False
        return is_a_tty() and not is_windows

    def _get_output_stream(self):
        return self._output_stream_factory.get_output_stream()



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr import emrutils, exceptions, helptext
from awscli.customizations.emr.command import Command


class ModifyClusterAttr(Command):
    NAME = 'modify-cluster-attributes'
    DESCRIPTION = (
        "Modifies the cluster attributes 'visible-to-all-users', "
        " 'termination-protected' and 'unhealthy-node-replacement'."
    )
    ARG_TABLE = [
        {
            'name': 'cluster-id',
            'required': True,
            'help_text': helptext.CLUSTER_ID,
        },
        {
            'name': 'visible-to-all-users',
            'required': False,
            'action': 'store_true',
            'group_name': 'visible',
            'help_text': helptext.VISIBILITY,
        },
        {
            'name': 'no-visible-to-all-users',
            'required': False,
            'action': 'store_true',
            'group_name': 'visible',
            'help_text': helptext.VISIBILITY,
        },
        {
            'name': 'termination-protected',
            'required': False,
            'action': 'store_true',
            'group_name': 'terminate',
            'help_text': 'Set termination protection on or off',
        },
        {
            'name': 'no-termination-protected',
            'required': False,
            'action': 'store_true',
            'group_name': 'terminate',
            'help_text': 'Set termination protection on or off',
        },
        {
            'name': 'auto-terminate',
            'required': False,
            'action': 'store_true',
            'group_name': 'auto_terminate',
            'help_text': 'Set cluster auto terminate after completing all the steps on or off',
        },
        {
            'name': 'no-auto-terminate',
            'required': False,
            'action': 'store_true',
            'group_name': 'auto_terminate',
            'help_text': 'Set cluster auto terminate after completing all the steps on or off',
        },
        {
            'name': 'unhealthy-node-replacement',
            'required': False,
            'action': 'store_true',
            'group_name': 'UnhealthyReplacement',
            'help_text': 'Set Unhealthy Node Replacement on or off',
        },
        {
            'name': 'no-unhealthy-node-replacement',
            'required': False,
            'action': 'store_true',
            'group_name': 'UnhealthyReplacement',
            'help_text': 'Set Unhealthy Node Replacement on or off',
        },
    ]

    def _run_main_command(self, args, parsed_globals):
        if args.visible_to_all_users and args.no_visible_to_all_users:
            raise exceptions.MutualExclusiveOptionError(
                option1='--visible-to-all-users',
                option2='--no-visible-to-all-users',
            )
        if args.termination_protected and args.no_termination_protected:
            raise exceptions.MutualExclusiveOptionError(
                option1='--termination-protected',
                option2='--no-termination-protected',
            )
        if args.auto_terminate and args.no_auto_terminate:
            raise exceptions.MutualExclusiveOptionError(
                option1='--auto-terminate', option2='--no-auto-terminate'
            )
        if (
            args.unhealthy_node_replacement
            and args.no_unhealthy_node_replacement
        ):
            raise exceptions.MutualExclusiveOptionError(
                option1='--unhealthy-node-replacement',
                option2='--no-unhealthy-node-replacement',
            )
        if not (
            args.termination_protected
            or args.no_termination_protected
            or args.visible_to_all_users
            or args.no_visible_to_all_users
            or args.auto_terminate
            or args.no_auto_terminate
            or args.unhealthy_node_replacement
            or args.no_unhealthy_node_replacement
        ):
            raise exceptions.MissingClusterAttributesError()

        if args.visible_to_all_users or args.no_visible_to_all_users:
            visible = (
                args.visible_to_all_users and not args.no_visible_to_all_users
            )
            parameters = {
                'JobFlowIds': [args.cluster_id],
                'VisibleToAllUsers': visible,
            }
            emrutils.call_and_display_response(
                self._session,
                'SetVisibleToAllUsers',
                parameters,
                parsed_globals,
            )

        if args.termination_protected or args.no_termination_protected:
            protected = (
                args.termination_protected
                and not args.no_termination_protected
            )
            parameters = {
                'JobFlowIds': [args.cluster_id],
                'TerminationProtected': protected,
            }
            emrutils.call_and_display_response(
                self._session,
                'SetTerminationProtection',
                parameters,
                parsed_globals,
            )

        if args.auto_terminate or args.no_auto_terminate:
            auto_terminate = args.auto_terminate and not args.no_auto_terminate
            parameters = {
                'JobFlowIds': [args.cluster_id],
                'KeepJobFlowAliveWhenNoSteps': not auto_terminate,
            }
            emrutils.call_and_display_response(
                self._session,
                'SetKeepJobFlowAliveWhenNoSteps',
                parameters,
                parsed_globals,
            )

        if (
            args.unhealthy_node_replacement
            or args.no_unhealthy_node_replacement
        ):
            protected = (
                args.unhealthy_node_replacement
                and not args.no_unhealthy_node_replacement
            )
            parameters = {
                'JobFlowIds': [args.cluster_id],
                'UnhealthyNodeReplacement': protected,
            }
            emrutils.call_and_display_response(
                self._session,
                'SetUnhealthyNodeReplacement',
                parameters,
                parsed_globals,
            )

        return 0



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
NOTE: All classes and functions in this module are considered private and are
subject to abrupt breaking changes. Please do not use them directly.
"""

from contextlib import contextmanager
from contextvars import ContextVar
from copy import deepcopy
from dataclasses import dataclass, field
from functools import wraps
from typing import Set


@dataclass
class ClientContext:
    """
    Encapsulation of objects tracked within the ``_context`` context variable.

    ``features`` is a set responsible for storing features used during
    preparation of an AWS request. ``botocore.useragent.register_feature_id``
    is used to add to this set.
    """

    features: Set[str] = field(default_factory=set)


_context = ContextVar("_context")


def get_context():
    """Get the current ``_context`` context variable if set, else None."""
    return _context.get(None)


def set_context(ctx):
    """Set the current ``_context`` context variable.

    :type ctx: ClientContext
    :param ctx: Client context object to set as the current context variable.

    :rtype: contextvars.Token
    :returns: Token object used to revert the context variable to what it was
        before the corresponding set.
    """
    token = _context.set(ctx)
    return token


def reset_context(token):
    """Reset the current ``_context`` context variable.

    :type token: contextvars.Token
    :param token: Token object to reset the context variable.
    """
    _context.reset(token)


@contextmanager
def start_as_current_context(ctx=None):
    """
    Context manager that copies the passed or current context object and sets
    it as the current context variable. If no context is found, a new
    ``ClientContext`` object is created. It mainly ensures the context variable
    is reset to the previous value once the executed code returns.

    Example usage:

        def my_feature():
            with start_as_current_context():
                register_feature_id('MY_FEATURE')
                pass

    :type ctx: ClientContext
    :param ctx: The client context object to set as the new context variable.
        If not provided, the current or a new context variable is used.
    """
    current = ctx or get_context()
    if current is None:
        new = ClientContext()
    else:
        new = deepcopy(current)
    token = set_context(new)
    try:
        yield
    finally:
        reset_context(token)


def with_current_context(hook=None):
    """
    Decorator that wraps ``start_as_current_context`` and optionally invokes a
    hook within the newly-set context. This is just syntactic sugar to avoid
    indenting existing code under the context manager.

    Example usage:

        @with_current_context(partial(register_feature_id, 'MY_FEATURE'))
        def my_feature():
            pass

    :type hook: callable
    :param hook: A callable that will be invoked within the scope of the
        ``start_as_current_context`` context manager.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with start_as_current_context():
                if hook:
                    hook()
                return func(*args, **kwargs)

        return wrapper

    return decorator



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# Class to create default roles for lifecycle
import logging

from awscli.clidriver import CLIOperationCaller
from awscli.customizations.commands import BasicCommand
from awscli.customizations.dlm.constants import (
    LIFECYCLE_DEFAULT_ROLE_ASSUME_POLICY,
    POLICY_ARN_PATTERN,
    RESOURCE_TYPE_IMAGE,
    RESOURCE_TYPE_SNAPSHOT,
    RESOURCES,
)
from awscli.customizations.dlm.iam import IAM
from awscli.customizations.exceptions import ConfigurationError
from awscli.customizations.utils import get_policy_arn_suffix

LOG = logging.getLogger(__name__)


def _construct_result(create_role_response, get_policy_response):
    get_policy_response.pop('ResponseMetadata', None)
    create_role_response.pop('ResponseMetadata', None)
    result = {'RolePolicy': get_policy_response}
    result.update(create_role_response)
    return result


# Display the result as formatted json
def display_response(session, operation_name, result, parsed_globals):
    if result is not None:
        cli_operation_caller = CLIOperationCaller(session)
        # Calling a private method. Should be changed after the functionality
        # is moved outside CliOperationCaller.
        cli_operation_caller._display_response(
            operation_name, result, parsed_globals
        )


# Get policy arn from region and policy name
def get_policy_arn(region, policy_name):
    region_suffix = get_policy_arn_suffix(region)
    role_arn = POLICY_ARN_PATTERN.format(region_suffix, policy_name)
    return role_arn


# Method to parse the arguments to get the region value
def get_region(session, parsed_globals):
    region = parsed_globals.region
    if region is None:
        region = session.get_config_variable('region')
    return region


class CreateDefaultRole(BasicCommand):
    NAME = "create-default-role"
    DESCRIPTION = (
        'Creates the default IAM role '
        ' which will be used by Lifecycle service.\n'
        'If the role does not exist, create-default-role '
        'will automatically create it and set its policy.'
        ' If the role has been already '
        'created, create-default-role'
        ' will not update its policy.'
        '\n'
    )
    ARG_TABLE = [
        {
            'name': 'iam-endpoint',
            'no_paramfile': True,
            'help_text': '<p>The IAM endpoint to call for creating the roles.'
            ' This is optional and should only be specified when a'
            ' custom endpoint should be called for IAM operations'
            '.</p>',
        },
        {
            'name': 'resource-type',
            'default': RESOURCE_TYPE_SNAPSHOT,
            'choices': [RESOURCE_TYPE_SNAPSHOT, RESOURCE_TYPE_IMAGE],
            'help_text': (
                "<p>The resource type for which the role needs to be created."
                " The available options are '%s' and '%s'."
                " This parameter defaults to '%s'.</p>"
                % (
                    RESOURCE_TYPE_SNAPSHOT,
                    RESOURCE_TYPE_IMAGE,
                    RESOURCE_TYPE_SNAPSHOT,
                )
            ),
        },
    ]

    def __init__(self, session):
        super(CreateDefaultRole, self).__init__(session)

    def _run_main(self, parsed_args, parsed_globals):
        """Call to run the commands"""

        self._region = get_region(self._session, parsed_globals)
        self._endpoint_url = parsed_args.iam_endpoint
        self._resource_type = parsed_args.resource_type
        self._iam_client = IAM(
            self._session.create_client(
                'iam',
                region_name=self._region,
                endpoint_url=self._endpoint_url,
                verify=parsed_globals.verify_ssl,
            )
        )

        result = self._create_default_role_if_not_exists(parsed_globals)

        display_response(self._session, 'create_role', result, parsed_globals)

        return 0

    def _create_default_role_if_not_exists(self, parsed_globals):
        """Method to create default lifecycle role
        if it doesn't exist already
        """

        role_name = RESOURCES[self._resource_type]['default_role_name']
        assume_role_policy = LIFECYCLE_DEFAULT_ROLE_ASSUME_POLICY

        if self._iam_client.check_if_role_exists(role_name):
            LOG.debug('Role %s exists', role_name)
            return None

        LOG.debug(
            'Role %s does not exist. Creating default role for Lifecycle',
            role_name,
        )

        # Get Region
        region = get_region(self._session, parsed_globals)

        if region is None:
            raise ConfigurationError(
                'You must specify a region. You can also configure your '
                'region by running "aws configure".'
            )

        managed_policy_arn = get_policy_arn(
            region, RESOURCES[self._resource_type]['default_policy_name']
        )

        # Don't proceed if managed policy does not exist
        if not self._iam_client.check_if_policy_exists(managed_policy_arn):
            LOG.debug('Managed Policy %s does not exist.', managed_policy_arn)
            return None

        LOG.debug('Managed Policy %s exists.', managed_policy_arn)
        # Create default role
        create_role_response = self._iam_client.create_role_with_trust_policy(
            role_name, assume_role_policy
        )
        # Attach policy to role
        self._iam_client.attach_policy_to_role(managed_policy_arn, role_name)

        # Construct result
        get_policy_response = self._iam_client.get_policy(managed_policy_arn)
        return _construct_result(create_role_response, get_policy_response)



import base64
import binascii
import hashlib
import json
import sys
from abc import ABC, abstractmethod
from os import path

from awscrt.crypto import RSA, RSASignatureAlgorithm

from awscli.customizations.cloudtrail.utils import (
    PublicKeyProvider,
    parse_date,
)
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError

SIGN_FILE_NAME = "result_sign.json"


def format_prefix(prefix):
    """
    Make prefix end with '/'

    :rtype: string
    :return: Return the prefix that end with '/'
    """
    if prefix == "" or prefix[-1] == "/":
        return prefix
    else:
        return prefix + "/"


class ValidationError(Exception):
    """Exception raised when something wrong in the validation process."""

    def __init__(self, message):
        message = f"ValidationError: {message}"
        super().__init__(message)


class InformationCollectionError(ValueError):
    """Exception raised when fail to collect required information."""

    def __init__(self, message):
        message = f"InformationCollectionError: {message}"
        super().__init__(message)


class Sha256RsaSignatureValidator:
    """
    Validates SHA256withRSA signed signature.
    """

    def validate(self, public_key_base64, sign_file):
        """Validates the signature with public key.

        Throws a ValidationError when the signature is invalid.

        :type public_key_base64: string
        :param public_key_base64: Public key bytes in base 64.
        :type sign_file: dict
        :param sign_file: Dict of sign file data returned when JSON
            decoding a manifest.
        """
        public_key = self._load_public_key(public_key_base64)
        signature_bytes = binascii.unhexlify(sign_file["hashSignature"])
        result = public_key.verify(
            signature_algorithm=RSASignatureAlgorithm.PKCS1_5_SHA256,
            digest=hashlib.sha256(
                self._create_string_to_sign(sign_file)
            ).digest(),
            signature=signature_bytes,
        )
        if not result:
            # The previous implementation caught a cryptography.exceptions.InvalidSignature
            # exception here and re-raised the DigestSignatureError to avoid the stack trace
            # leaking any information about the key.
            raise ValidationError("Invalid signature in sign file")

    def _create_string_to_sign(self, sign_file):
        hash_list = []
        for file_info in sign_file["files"]:
            hash_list.append(file_info["fileHashValue"])

        string_to_sign = " ".join(hash_list)
        return string_to_sign.encode()

    def _load_public_key(self, public_key_base64):
        try:
            decoded_key = base64.b64decode(public_key_base64)
            return RSA.new_public_key_from_der_data(decoded_key)
        except ValueError:
            raise ValidationError(
                f"Sign file invalid, unable to load PKCS #1 key: {public_key_base64}"
            )


class BaseSignFileProvider(ABC):
    @abstractmethod
    def provide_sign_file(self):
        """
        Get the sign file and query execute time.

        :rtype: string, string
        :return: Return the sign file and dateTime of the query export time.
        """
        pass

    def _load_sign_file(self, sign_file_str):
        try:
            return json.loads(sign_file_str)
        except json.JSONDecodeError as e:
            raise InformationCollectionError(
                f"Unable to load result_sign.json file, due to {str(e)}"
            )


class LocalSignFileProvider(BaseSignFileProvider):
    def __init__(
        self,
        local_path_prefix=None,
    ):
        self._local_path_prefix = local_path_prefix

    def provide_sign_file(self):
        with open(path.join(self._local_path_prefix, SIGN_FILE_NAME)) as f:
            sign_file = self._load_sign_file(f.read())

        return sign_file


class S3SignFileProvider(BaseSignFileProvider):
    def __init__(
        self,
        s3_client=None,
        s3_bucket=None,
        s3_path_prefix=None,
    ):
        self._s3_client = s3_client
        self._s3_bucket = s3_bucket
        self._s3_path_prefix = s3_path_prefix

    def provide_sign_file(self):
        sign_file_byte = self._s3_client.get_object(
            Bucket=self._s3_bucket, Key=self._s3_path_prefix + SIGN_FILE_NAME
        )["Body"].read()
        sign_file = self._load_sign_file(sign_file_byte.decode("utf-8"))

        return sign_file


class BaseExportFilesHashValidator(ABC):
    @abstractmethod
    def validate_export_files(self, sign_file):
        """Validates all the export files in the query.

        Throws a ValidationError when export file's hash value is inconsistent
        with the hash value recorded in sign file.

        :type sign_file: dict
        :param sign_file: Dict of sign file data returned when JSON
            decoding a manifest.
        """
        pass

    def _validate_hash_value(self, read_stream, file_name, expected_hash):
        rolling_hash = hashlib.sha256()
        for chunk in iter(lambda: read_stream.read(2048), b""):
            rolling_hash.update(chunk)
        computed_hash = rolling_hash.hexdigest()
        if computed_hash != expected_hash:
            raise ValidationError(
                f"File {file_name} has inconsistent hash value with hash value recorded"
                f" in sign file, hash value in sign file is {expected_hash} ,"
                f" but get {computed_hash}"
            )

    def _validate_number_of_files(self, sign_file):
        try:
            if len(sign_file["files"]) == 0:
                raise ValidationError("No export file was found in sign file.")
        except TypeError:
            raise ValidationError("Invalid sign file provided.")


class S3ExportFilesHashValidator(BaseExportFilesHashValidator):
    def __init__(
        self,
        s3_client=None,
        s3_bucket=None,
        s3_path_prefix=None,
    ):
        self._s3_client = s3_client
        self._s3_bucket = s3_bucket
        self._s3_path_prefix = s3_path_prefix

    def validate_export_files(self, sign_file):
        self._validate_number_of_files(sign_file)

        for file_info in sign_file["files"]:
            key = self._s3_path_prefix + file_info["fileName"]
            response = self._s3_client.get_object(
                Bucket=self._s3_bucket, Key=key
            )
            self._validate_hash_value(
                response["Body"],
                file_info["fileName"],
                file_info["fileHashValue"],
            )


class LocalExportFilesHashValidator(BaseExportFilesHashValidator):
    def __init__(
        self,
        local_path_prefix=None,
    ):
        self.local_path_prefix = local_path_prefix

    def validate_export_files(self, sign_file):
        self._validate_number_of_files(sign_file)

        for file_info in sign_file["files"]:
            with open(
                path.join(self.local_path_prefix, file_info["fileName"]), "rb"
            ) as export_file:
                self._validate_hash_value(
                    export_file,
                    file_info["fileName"],
                    file_info["fileHashValue"],
                )


class CloudTrailVerifyQueryResult(BasicCommand):
    """
    Validates export files output from a CloudTrail Lake query.
    """

    NAME = "verify-query-results"
    DESCRIPTION = """
    Validates CloudTrail Lake query's export files.

    This command uses the query export and sign file delivered to you to
    perform the validation.

    The AWS CLI allows you to detect the following types of changes:

    - Modification or deletion of CloudTrail Lake query's export files.

    To validate export files with the AWS CLI, the following preconditions must
    be met:

    - You must have online connectivity to AWS.
    - You must put the sign file and export file in the specified path prefix
    - You must not rename the delivered export file and sign file
    - For validate export files from S3: (1) You must have read access to the
      S3 bucket that contains the sign and export file. (2) The digest and log
      files must not have been moved from the original S3 location where
      CloudTrail delivered them.

    .. note::
        For verify export file from S3, this command requires that the user or
        role executing the command has permission to call GetObject, and
        GetBucketLocation for the bucket that store the export file.
    """

    ARG_TABLE = [
        {
            "name": "local-export-path",
            "required": False,
            "cli_type_name": "string",
            "help_text": "Specifies the local directory of export and sign file,"
            " e.g. /local/path/to/export/file/ ",
        },
        {
            "name": "s3-bucket",
            "required": False,
            "cli_type_name": "string",
            "help_text": "Specifies the S3 bucket name that store the query "
            "result and sign file "
            "This parameter can not coexist with local-export-path.",
        },
        {
            "name": "s3-prefix",
            "required": False,
            "cli_type_name": "string",
            "help_text": "Specifies the S3 path of the S3 folder that contain"
            "export and sign file, e.g. bucket_name/s3/path/ . "
            "This parameter can not coexist with local-export-path. If the "
            "files located in s3 bucket root directory, then no need to "
            "provide this parameter.",
        },
    ]

    def __init__(self, session):
        super().__init__(session)
        self._local_export_path = None
        self._s3_prefix = None
        self._s3_bucket = None
        self._is_from_s3 = None
        self._return_code = 255

        self._s3_client_provider = None
        self._cloudtrail_client = None
        self._source_region = None

    def _run_main(self, args, parsed_globals):
        self._call(args, parsed_globals)
        return self._return_code

    def handle_args(self, args):
        if not args.local_export_path and not args.s3_bucket:
            raise ParamValidationError(
                "Require parameter --s3-bucket or --local-export-path."
            )
        if args.local_export_path and (args.s3_bucket or args.s3_prefix):
            raise ParamValidationError(
                "Parameter --local-export-path can not be specified with"
                " parameter --s3-bucket nor --s3-prefix."
            )

        if args.local_export_path:
            self._is_from_s3 = False
            self._local_export_path = args.local_export_path
        else:
            self._is_from_s3 = True
            self._s3_prefix = (
                "" if args.s3_prefix is None else format_prefix(args.s3_prefix)
            )
            self._s3_bucket = args.s3_bucket

    def setup_services(self, parsed_globals):
        self._source_region = parsed_globals.region
        client_args = {
            "region_name": parsed_globals.region,
            "verify": parsed_globals.verify_ssl,
        }

        if parsed_globals.endpoint_url is not None:
            client_args["endpoint_url"] = parsed_globals.endpoint_url
        self._cloudtrail_client = self._session.create_client(
            "cloudtrail", **client_args
        )

    def _call(self, args, parsed_globals):
        self.handle_args(args)
        self.setup_services(parsed_globals)
        (
            public_key_provider,
            signature_validator,
            sign_file_provider,
            hash_validator,
        ) = self._initialize_components(
            self._is_from_s3,
            self._cloudtrail_client,
            self._s3_bucket,
            self._s3_prefix,
            self._local_export_path,
        )

        sign_file = sign_file_provider.provide_sign_file()
        hash_validator.validate_export_files(sign_file)

        public_key = public_key_provider.get_public_key(
            parse_date(sign_file["queryCompleteTime"]),
            sign_file["publicKeyFingerprint"],
        )
        signature_validator.validate(public_key, sign_file)
        self._return_code = 0
        sys.stdout.write(
            "Successfully validated sign and query result files\n"
        )

    def _initialize_components(
        self,
        is_from_s3,
        cloudtrail_client,
        s3_bucket=None,
        s3_prefix=None,
        local_export_path=None,
    ):
        public_key_provider = PublicKeyProvider(cloudtrail_client)
        signature_validator = Sha256RsaSignatureValidator()
        if is_from_s3:
            s3_client = self._session.create_client("s3", self._source_region)
            sign_file_provider = S3SignFileProvider(
                s3_client=s3_client,
                s3_bucket=s3_bucket,
                s3_path_prefix=s3_prefix,
            )
            export_file_hash_validator = S3ExportFilesHashValidator(
                s3_client=s3_client,
                s3_bucket=s3_bucket,
                s3_path_prefix=s3_prefix,
            )
        else:
            sign_file_provider = LocalSignFileProvider(
                local_path_prefix=local_export_path
            )
            export_file_hash_validator = LocalExportFilesHashValidator(
                local_path_prefix=local_export_path
            )

        return (
            public_key_provider,
            signature_validator,
            sign_file_provider,
            export_file_hash_validator,
        )



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import base64
import binascii
import hashlib
import json
import logging
import re
import sys
import zlib
from datetime import datetime, timedelta
from zlib import error as ZLibError

from awscrt.crypto import RSA, RSASignatureAlgorithm
from botocore.exceptions import ClientError
from dateutil import parser, tz

from awscli.customizations.cloudtrail.utils import (
    get_account_id_from_arn,
    get_trail_by_arn,
)
from awscli.customizations.commands import BasicCommand
from awscli.schema import ParameterRequiredError

LOG = logging.getLogger(__name__)
DATE_FORMAT = '%Y%m%dT%H%M%SZ'
DISPLAY_DATE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


def format_date(date):
    """Returns a formatted date string in a CloudTrail date format"""
    return date.strftime(DATE_FORMAT)


def format_display_date(date):
    """Returns a formatted date string meant for CLI output"""
    return date.strftime(DISPLAY_DATE_FORMAT)


def normalize_date(date):
    """Returns a normalized date using a UTC timezone"""
    return date.replace(tzinfo=tz.tzutc())


def is_backfill_digest_key(digest_key):
    """Utility function to determine if a digest key represents a backfill digest file"""
    return digest_key.endswith('_backfill.json.gz')


def extract_digest_key_date(digest_s3_key):
    """Extract the timestamp portion of a manifest file.

    Manifest file names take the following form:
    AWSLogs/{account}/CloudTrail-Digest/{region}/{ymd}/{account}_CloudTrail \
    -Digest_{region}_{name}_region_{date}.json.gz

    For backfill files:
    AWSLogs/{account}/CloudTrail-Digest/{region}/{ymd}/{account}_CloudTrail \
    -Digest_{region}_{name}_region_{date}_backfill.json.gz
    """
    if is_backfill_digest_key(digest_s3_key):
        # Backfill files have _backfill suffix before .json.gz
        return digest_s3_key[-33:-17]
    else:
        # Regular digest files
        return digest_s3_key[-24:-8]


def parse_date(date_string):
    try:
        return parser.parse(date_string)
    except ValueError:
        raise ValueError(f'Unable to parse date value: {date_string}')


def assert_cloudtrail_arn_is_valid(trail_arn):
    """Ensures that the arn looks correct.

    ARNs look like: arn:aws:cloudtrail:us-east-1:123456789012:trail/foo"""
    pattern = re.compile(r'arn:.+:cloudtrail:.+:\d{12}:trail/.+')
    if not pattern.match(trail_arn):
        raise ValueError(f'Invalid trail ARN provided: {trail_arn}')


def create_digest_traverser(
    cloudtrail_client,
    organization_client,
    s3_client_provider,
    trail_arn,
    trail_source_region=None,
    on_invalid=None,
    on_gap=None,
    on_missing=None,
    bucket=None,
    prefix=None,
    account_id=None,
):
    """Creates a CloudTrail DigestTraverser and its object graph.

    :type cloudtrail_client: botocore.client.CloudTrail
    :param cloudtrail_client: Client used to connect to CloudTrail
    :type organization_client: botocore.client.organizations
    :param organization_client: Client used to connect to Organizations
    :type s3_client_provider: S3ClientProvider
    :param s3_client_provider: Used to create Amazon S3 client per/region.
    :param trail_arn: CloudTrail trail ARN
    :param trail_source_region: The scanned region of a trail.
    :param on_invalid: Callback that is invoked when validating a digest fails.
    :param on_gap: Callback that is invoked when a digest has no link to the
        previous digest, but there are more digests to validate. This can
        happen when a trail is disabled for a period of time.
    :param on_missing: Callback that is invoked when a digest file has been
        deleted from Amazon S3 but is supposed to be present.
    :param bucket: Amazon S3 bucket of the trail if it is different than the
        bucket that is currently associated with the trail.
    :param prefix: bucket: Key prefix prepended to each digest and log placed
        in the Amazon S3 bucket if it is different than the prefix that is
        currently associated with the trail.
    :param account_id: The account id for which the digest files are
        validated. For normal trails this is the caller account, for
        organization trails it is the member account.

    ``on_gap``, ``on_invalid``, and ``on_missing`` callbacks are invoked with
    the following named arguments:

    - ``bucket`: The next S3 bucket.
    - ``next_key``: (optional) Next digest key that was found in the bucket.
    - ``next_end_date``: (optional) End date of the next found digest.
    - ``last_key``: The last digest key that was found.
    - ``last_start_date``: (optional) Start date of last found digest.
    - ``message``: (optional) Message string about the notification.
    """
    assert_cloudtrail_arn_is_valid(trail_arn)
    organization_id = None
    if bucket is None:
        # Determine the bucket and prefix based on the trail arn.
        trail_info = get_trail_by_arn(cloudtrail_client, trail_arn)
        LOG.debug(f'Loaded trail info: {trail_info}')
        bucket = trail_info['S3BucketName']
        prefix = trail_info.get('S3KeyPrefix', None)
        is_org_trail = trail_info.get('IsOrganizationTrail')
        if is_org_trail:
            if not account_id:
                raise ParameterRequiredError(
                    "Missing required parameter for organization "
                    "trail: '--account-id'"
                )
            organization_id = organization_client.describe_organization()[
                'Organization'
            ]['Id']

    # Determine the region from the ARN (e.g., arn:aws:cloudtrail:REGION:...)
    trail_region = trail_arn.split(':')[3]
    # Determine the name from the ARN (the last part after "/")
    trail_name = trail_arn.split('/')[-1]
    # If account id is not specified parse it from trail ARN
    if not account_id:
        account_id = get_account_id_from_arn(trail_arn)

    digest_provider = DigestProvider(
        account_id=account_id,
        trail_name=trail_name,
        s3_client_provider=s3_client_provider,
        trail_source_region=trail_source_region,
        trail_home_region=trail_region,
        organization_id=organization_id,
    )
    return DigestTraverser(
        digest_provider=digest_provider,
        starting_bucket=bucket,
        starting_prefix=prefix,
        on_invalid=on_invalid,
        on_gap=on_gap,
        on_missing=on_missing,
        public_key_provider=PublicKeyProvider(cloudtrail_client),
    )


class S3ClientProvider:
    """Creates Amazon S3 clients and determines the region name of a client.

    This class will cache the location constraints of previously requested
    buckets and cache previously created clients for the same region.
    """

    def __init__(self, session, get_bucket_location_region='us-east-1'):
        self._session = session
        self._get_bucket_location_region = get_bucket_location_region
        self._client_cache = {}
        self._region_cache = {}

    def get_client(self, bucket_name):
        """Creates an S3 client that can work with the given bucket name"""
        region_name = self._get_bucket_region(bucket_name)
        return self._create_client(region_name)

    def _get_bucket_region(self, bucket_name):
        """Returns the region of a bucket"""
        if bucket_name not in self._region_cache:
            client = self._create_client(self._get_bucket_location_region)
            result = client.get_bucket_location(Bucket=bucket_name)
            region = result['LocationConstraint'] or 'us-east-1'
            self._region_cache[bucket_name] = region
        return self._region_cache[bucket_name]

    def _create_client(self, region_name):
        """Creates an Amazon S3 client for the given region name"""
        if region_name not in self._client_cache:
            client = client = self._session.create_client(
                's3', region_name=region_name
            )
            # Remove the CLI error event that prevents exceptions.
            self._client_cache[region_name] = client
        return self._client_cache[region_name]


class DigestError(ValueError):
    """Exception raised when a digest fails to validate"""

    pass


class DigestSignatureError(DigestError):
    """Exception raised when a digest signature is invalid"""

    def __init__(self, bucket, key):
        message = (
            f'Digest file\ts3://{bucket}/{key}\tINVALID: signature verification '
            'failed'
        )
        super().__init__(message)


class InvalidDigestFormat(DigestError):
    """Exception raised when a digest has an invalid format"""

    def __init__(self, bucket, key):
        message = f'Digest file\ts3://{bucket}/{key}\tINVALID: invalid format'
        super().__init__(message)


class PublicKeyProvider:
    """Retrieves public keys from CloudTrail within a date range."""

    def __init__(self, cloudtrail_client):
        self._cloudtrail_client = cloudtrail_client

    def get_public_keys(self, start_date, end_date):
        """Loads public keys in a date range into a returned dict.

        :type start_date: datetime
        :param start_date: Start date of a date range.
        :type end_date: datetime
        :param end_date: End date of a date range.
        :rtype: dict
        :return: Returns a dict where each key is the fingerprint of the
            public key, and each value is a dict of public key data.
        """
        public_keys = self._cloudtrail_client.list_public_keys(
            StartTime=start_date, EndTime=end_date
        )
        public_keys_in_range = public_keys['PublicKeyList']
        LOG.debug(f'Loaded public keys in range: {public_keys_in_range}')
        return dict((key['Fingerprint'], key) for key in public_keys_in_range)


class DigestProvider:
    """
    Retrieves digest keys and digests from Amazon S3.

    This class is responsible for determining the full list of digest files
    in a bucket and loading digests from the bucket into a JSON decoded
    dict. This class is not responsible for validation or iterating from
    one digest to the next.
    """

    def __init__(
        self,
        s3_client_provider,
        account_id,
        trail_name,
        trail_home_region,
        trail_source_region=None,
        organization_id=None,
    ):
        self._client_provider = s3_client_provider
        self.trail_name = trail_name
        self.account_id = account_id
        self.trail_home_region = trail_home_region
        self.trail_source_region = trail_source_region or trail_home_region
        self.organization_id = organization_id
        self._digest_cache = {}

    def load_all_digest_keys_in_range(
        self, bucket, prefix, start_date, end_date
    ):
        """Load all digest keys and separate into standard and backfill lists.

        Performs a single S3 list operation and separates keys into standard
        and backfill digest lists during iteration for optimal performance.

        :param bucket: S3 bucket name
        :param prefix: S3 key prefix
        :param start_date: Start date for digest range
        :param end_date: End date for digest range
        :return: Tuple of (standard_digests, backfill_digests) lists
        :rtype: tuple
        """
        standard_digests = []
        backfill_digests = []
        marker = self._create_digest_key(start_date, prefix)
        s3_digest_files_prefix = self._create_digest_prefix(start_date, prefix)
        client = self._client_provider.get_client(bucket)
        paginator = client.get_paginator('list_objects')
        page_iterator = paginator.paginate(
            Bucket=bucket, Marker=marker, Prefix=s3_digest_files_prefix
        )
        key_filter = page_iterator.search('Contents[*].Key')
        # Create a target start end end date
        target_start_date = format_date(normalize_date(start_date))
        # Add one hour to the end_date to get logs that spilled over to next.
        target_end_date = format_date(
            normalize_date(end_date + timedelta(hours=1))
        )
        # Ensure digests are from the same trail.
        digest_key_regex = re.compile(self._create_digest_key_regex(prefix))
        for key in key_filter:
            if not (key and digest_key_regex.match(key)):
                continue
            # Use a lexicographic comparison to know when to stop.
            extracted_date = extract_digest_key_date(key)
            if extracted_date > target_end_date:
                break
            # Only append digests after the start date.
            if extracted_date < target_start_date:
                continue
            if is_backfill_digest_key(key):
                backfill_digests.append(key)
            else:
                standard_digests.append(key)
        return standard_digests, backfill_digests

    def load_digest_keys_in_range(
        self, bucket, prefix, start_date, end_date, is_backfill=False
    ):
        """Returns a list of digest keys in the date range.

        This method uses caching to avoid duplicate S3 list operations.
        On first call, it loads all digest keys and caches them separated
        by type. Subsequent calls return the appropriate cached list.

        :param bucket: S3 bucket name
        :param prefix: S3 key prefix
        :param start_date: Start date for digest range
        :param end_date: End date for digest range
        :param is_backfill: Optional filter - True for backfill digests only,
                           False for standard digests only
        :return: List of digest keys matching the specified type
        :rtype: list
        """
        cache_key = (bucket, prefix, start_date, end_date)

        if cache_key not in self._digest_cache:
            standard_digests, backfill_digests = (
                self.load_all_digest_keys_in_range(
                    bucket, prefix, start_date, end_date
                )
            )
            self._digest_cache[cache_key] = {
                'standard': standard_digests,
                'backfill': backfill_digests,
            }

        if is_backfill:
            return self._digest_cache[cache_key]['backfill']
        else:
            return self._digest_cache[cache_key]['standard']

    def fetch_digest(self, bucket, key):
        """Loads a digest by key from S3.

        Returns the JSON decode data and GZIP inflated raw content.
        For backfill digests, also extracts the backfill-generation-timestamp.
        """
        client = self._client_provider.get_client(bucket)
        result = client.get_object(Bucket=bucket, Key=key)
        try:
            digest = zlib.decompress(
                result['Body'].read(), zlib.MAX_WBITS | 16
            )
            digest_data = json.loads(digest.decode())
        except (ValueError, ZLibError):
            # Cannot gzip decode or JSON parse.
            raise InvalidDigestFormat(bucket, key)
        # Add the expected digest signature and algorithm to the dict.
        if (
            'signature' not in result['Metadata']
            or 'signature-algorithm' not in result['Metadata']
        ):
            raise DigestSignatureError(bucket, key)
        digest_data['_signature'] = result['Metadata']['signature']
        digest_data['_signature_algorithm'] = result['Metadata'][
            'signature-algorithm'
        ]

        if is_backfill_digest_key(key):
            if 'backfill-generation-timestamp' in result['Metadata']:
                digest_data['_backfill_generation_timestamp'] = result[
                    'Metadata'
                ]['backfill-generation-timestamp']
            else:
                raise InvalidDigestFormat(bucket, key)

        return digest_data, digest

    def _create_digest_key(self, start_date, key_prefix):
        """Computes an Amazon S3 key based on the provided data.

        The computed is what would have been placed in the S3 bucket if
        a log digest were created at a specific time. This computed key
        does not have to actually exist as it will only be used to as
        a Marker parameter in a list_objects call.

        :return: Returns a computed key as a string.
        """
        # Subtract one minute to ensure the dates are inclusive.
        date = start_date - timedelta(minutes=1)
        account_id = self.account_id
        date_str = format_date(date)
        ymd = date.strftime('%Y/%m/%d')
        source_region = self.trail_source_region
        home_region = self.trail_home_region
        name = self.trail_name

        if self.organization_id:
            organization_id = self.organization_id
            key = (
                f'AWSLogs/{organization_id}/{account_id}/CloudTrail-Digest/'
                f'{source_region}/{ymd}/{account_id}_CloudTrail-Digest_'
                f'{source_region}_{name}_{home_region}_{date_str}.json.gz'
            )
        else:
            key = (
                f'AWSLogs/{account_id}/CloudTrail-Digest/{source_region}/'
                f'{ymd}/{account_id}_CloudTrail-Digest_{source_region}_{name}_'
                f'{home_region}_{date_str}.json.gz'
            )

        if key_prefix:
            key = key_prefix + '/' + key
        return key

    def _create_digest_prefix(self, start_date, key_prefix):
        """Creates an S3 prefix to scope listing to trail's region.

        :return: Returns a prefix string to limit S3 listing scope.
        """
        template = 'AWSLogs/'
        template_params = {
            'account_id': self.account_id,
            'source_region': self.trail_source_region,
        }
        if self.organization_id:
            template += '{organization_id}/'
            template_params['organization_id'] = self.organization_id
        template += '{account_id}/CloudTrail-Digest/{source_region}'
        prefix = template.format(**template_params)
        if key_prefix:
            prefix = key_prefix + '/' + prefix
        return prefix

    def _create_digest_key_regex(self, key_prefix):
        """Creates a regular expression used to match against S3 keys for both standard and backfill digests"""
        account_id = re.escape(self.account_id)
        source_region = re.escape(self.trail_source_region)
        home_region = re.escape(self.trail_home_region)
        name = re.escape(self.trail_name)

        if self.organization_id:
            organization_id = self.organization_id
            key = (
                f'AWSLogs/{organization_id}/{account_id}/CloudTrail\\-Digest/'
                f'{source_region}/\\d+/\\d+/\\d+/{account_id}_CloudTrail\\-Digest_'
                f'{source_region}_{name}_{home_region}_.+(?:_backfill)?\\.json\\.gz'
            )
        else:
            key = (
                f'AWSLogs/{account_id}/CloudTrail\\-Digest/{source_region}/'
                f'\\d+/\\d+/\\d+/{account_id}_CloudTrail\\-Digest_'
                f'{source_region}_{name}_{home_region}_.+(?:_backfill)?\\.json\\.gz'
            )

        if key_prefix:
            key = re.escape(key_prefix) + '/' + key
        return '^' + key + '$'


class DigestTraverser:
    """Retrieves and validates digests within a date range."""

    # These keys are required to be present before validating the contents
    # of a digest.
    required_digest_keys = [
        'digestPublicKeyFingerprint',
        'digestS3Bucket',
        'digestS3Object',
        'previousDigestSignature',
        'digestEndTime',
        'digestStartTime',
    ]

    def __init__(
        self,
        digest_provider,
        starting_bucket,
        starting_prefix,
        public_key_provider,
        digest_validator=None,
        on_invalid=None,
        on_gap=None,
        on_missing=None,
    ):
        """
        :type digest_provider: DigestProvider
        :param digest_provider: DigestProvider object
        :param starting_bucket: S3 bucket where the digests are stored.
        :param starting_prefix: An optional prefix applied to each S3 key.
        :param public_key_provider: Provides public keys for a range.
        :param digest_validator: Validates digest using a validate method.
        :param on_invalid: Callback invoked when a digest is invalid.
        :param on_gap: Callback invoked when a digest has no parent, but
            there are still more digests to validate.
        :param on_missing: Callback invoked when a digest file is missing.
        """
        self.starting_bucket = starting_bucket
        self.starting_prefix = starting_prefix
        self.digest_provider = digest_provider
        self._public_key_provider = public_key_provider
        self._on_gap = on_gap
        self._on_invalid = on_invalid
        self._on_missing = on_missing
        if digest_validator is None:
            digest_validator = Sha256RSADigestValidator()
        self._digest_validator = digest_validator

    def traverse_digests(self, start_date, end_date=None, is_backfill=False):
        """Creates and returns a generator that yields validated digest data.

        Each yielded digest dictionary contains information about the digest
        and the log file associated with the digest. Digest files are validated
        before they are yielded. Whether or not the digest is successfully
        validated is stated in the "isValid" key value pair of the yielded
        dictionary.

        :type start_date: datetime
        :param start_date: Date to start validating from (inclusive).
        :type end_date: datetime
        :param end_date: Date to stop validating at (inclusive).
        :type is_backfill: bool
        :param is_backfill: Flag indicating whether to process backfill digests only.
        """
        if end_date is None:
            end_date = datetime.utcnow()
        end_date = normalize_date(end_date)
        start_date = normalize_date(start_date)
        bucket = self.starting_bucket
        prefix = self.starting_prefix

        digests = self._load_digests(
            bucket, prefix, start_date, end_date, is_backfill=is_backfill
        )

        # For regular digests, pre-load public keys. For backfill, start with empty dict
        public_keys = (
            {}
            if is_backfill
            else self._load_public_keys(
                start_date, end_date + timedelta(hours=2)
            )
        )

        yield from self._traverse_digest_chain(
            digests,
            bucket,
            prefix,
            start_date,
            public_keys,
            is_backfill=is_backfill,
        )

    def _traverse_digest_chain(
        self,
        digests,
        bucket,
        prefix,
        start_date,
        public_keys,
        is_backfill=False,
    ):
        """Traverses a single chain of digests

        :param is_backfill: Boolean indicating whether this chain contains backfill digests
        """
        key, end_date = self._get_last_digest(digests)
        last_start_date = end_date

        while key and start_date <= last_start_date:
            try:
                digest, end_date = self._load_and_validate_digest(
                    public_keys, bucket, key, is_backfill=is_backfill
                )
                last_start_date = normalize_date(
                    parse_date(digest['digestStartTime'])
                )
                previous_bucket = digest.get('previousDigestS3Bucket', None)
                previous_key = digest.get('previousDigestS3Object', None)
                yield digest
                if previous_bucket is None or previous_key is None:
                    # The chain is broken, so find next in digest store.
                    key, end_date = self._find_next_digest(
                        digests=digests,
                        bucket=bucket,
                        last_key=key,
                        last_start_date=last_start_date,
                        cb=self._on_gap,
                        is_cb_conditional=True,
                        is_backfill=is_backfill,
                    )
                else:
                    key = previous_key
                    if previous_bucket != bucket:
                        bucket = previous_bucket
                        # The bucket changed so reload the digest list.
                        digests = self._load_digests(
                            bucket,
                            prefix,
                            start_date,
                            end_date,
                            is_backfill=is_backfill,
                        )
            except ClientError as e:
                if e.response['Error']['Code'] != 'NoSuchKey':
                    raise e
                key, end_date = self._find_next_digest(
                    digests=digests,
                    bucket=bucket,
                    last_key=key,
                    last_start_date=last_start_date,
                    cb=self._on_missing,
                    message=str(e),
                    is_backfill=is_backfill,
                )
            except DigestError as e:
                key, end_date = self._find_next_digest(
                    digests=digests,
                    bucket=bucket,
                    last_key=key,
                    last_start_date=last_start_date,
                    cb=self._on_invalid,
                    message=str(e),
                    is_backfill=is_backfill,
                )
            except Exception as e:
                # Any other unexpected errors.
                key, end_date = self._find_next_digest(
                    digests=digests,
                    bucket=bucket,
                    last_key=key,
                    last_start_date=last_start_date,
                    cb=self._on_invalid,
                    message=f'Digest file\ts3://{bucket}/{key}\tINVALID: {str(e)}',
                    is_backfill=is_backfill,
                )

    def _load_digests(
        self, bucket, prefix, start_date, end_date, is_backfill=False
    ):
        return self.digest_provider.load_digest_keys_in_range(
            bucket=bucket,
            prefix=prefix,
            start_date=start_date,
            end_date=end_date,
            is_backfill=is_backfill,
        )

    def _find_next_digest(
        self,
        digests,
        bucket,
        last_key,
        last_start_date,
        cb=None,
        is_cb_conditional=False,
        message=None,
        is_backfill=False,
    ):
        """Finds the next digest in the bucket and invokes any callback."""
        next_key, next_end_date = self._get_last_digest(digests, last_key)
        if cb and (not is_cb_conditional or next_key):
            cb(
                bucket=bucket,
                next_key=next_key,
                last_key=last_key,
                next_end_date=next_end_date,
                last_start_date=last_start_date,
                message=message,
                is_backfill=is_backfill,
            )
        return next_key, next_end_date

    def _get_last_digest(self, digests, before_key=None):
        """Finds the previous digest key (either the last or before before_key)

        If no key is provided, the last digest is used. If a digest is found,
        the end date of the provider is adjusted to match the found key's end
        date.
        """
        if not digests:
            return None, None
        elif before_key is None:
            next_key = digests.pop()
            next_key_date = normalize_date(
                parse_date(extract_digest_key_date(next_key))
            )
            return next_key, next_key_date
        # find a key before the given key.
        before_key_date = parse_date(extract_digest_key_date(before_key))
        while digests:
            next_key = digests.pop()
            next_key_date = normalize_date(
                parse_date(extract_digest_key_date(next_key))
            )
            if next_key_date < before_key_date:
                LOG.debug(f"Next found key: {next_key}")
                return next_key, next_key_date
        return None, None

    def _load_and_validate_digest(
        self, public_keys, bucket, key, is_backfill=False
    ):
        """Loads and validates a digest from S3.

        :param public_keys: Public key dictionary of fingerprint to dict.
        :param bucket: S3 bucket name
        :param key: S3 key for the digest file
        :param is_backfill: Flag indicating if this is a backfill digest
        :return: Returns a tuple of the digest data as a dict and end_date
        :rtype: tuple
        """
        digest_data, digest = self.digest_provider.fetch_digest(bucket, key)

        # Validate required keys are present
        for required_key in self.required_digest_keys:
            if required_key not in digest_data:
                raise InvalidDigestFormat(bucket, key)

        # Ensure the bucket and key are the same as what's expected
        if (
            digest_data['digestS3Bucket'] != bucket
            or digest_data['digestS3Object'] != key
        ):
            raise DigestError(
                f'Digest file\ts3://{bucket}/{key}\tINVALID: has been moved from its '
                'original location'
            )

        fingerprint = digest_data['digestPublicKeyFingerprint']
        if fingerprint not in public_keys and is_backfill:
            # Backfill-specific logic to fetch public keys
            backfill_timestamp = normalize_date(
                parse_date(digest_data['_backfill_generation_timestamp'])
            )
            start_time = backfill_timestamp - timedelta(hours=1)
            end_time = backfill_timestamp + timedelta(hours=1)
            public_keys.update(self._load_public_keys(start_time, end_time))

        if fingerprint not in public_keys:
            error_message = (
                f'Digest file\ts3://{bucket}/{key}\tINVALID: public key not found in '
                f'region {self.digest_provider.trail_home_region} for fingerprint {fingerprint}'
            )
            raise DigestError(error_message)

        public_key_hex = public_keys[fingerprint]['Value']
        self._digest_validator.validate(
            bucket, key, public_key_hex, digest_data, digest
        )

        end_date = normalize_date(parse_date(digest_data['digestEndTime']))
        return digest_data, end_date

    def _load_public_keys(self, start_date, end_date):
        public_keys = self._public_key_provider.get_public_keys(
            start_date, end_date
        )
        if not public_keys:
            raise RuntimeError(
                f'No public keys found between {format_display_date(start_date)} and {format_display_date(end_date)}'
            )
        return public_keys


class Sha256RSADigestValidator:
    """
    Validates SHA256withRSA signed digests.

    The result of validating the digest is inserted into the digest_data
    dictionary using the isValid key value pair.
    """

    def validate(self, bucket, key, public_key, digest_data, inflated_digest):
        """Validates a digest file.

        Throws a DigestError when the digest is invalid.

        :param bucket: Bucket of the digest file
        :param key: Key of the digest file
        :param public_key: Public key bytes.
        :param digest_data: Dict of digest data returned when JSON
            decoding a manifest.
        :param inflated_digest: Inflated digest file contents as bytes.
        """
        try:
            decoded_key = base64.b64decode(public_key)
            public_key = RSA.new_public_key_from_der_data(decoded_key)
        except RuntimeError:
            raise DigestError(
                (
                    'Digest file\ts3://{}/{}\tINVALID: Unable to load PKCS #1 key'
                    ' with fingerprint {}'
                ).format(
                    bucket, key, digest_data['digestPublicKeyFingerprint']
                )
            )

        to_sign = self._create_string_to_sign(digest_data, inflated_digest)
        signature_bytes = binascii.unhexlify(digest_data['_signature'])

        result = public_key.verify(
            signature_algorithm=RSASignatureAlgorithm.PKCS1_5_SHA256,
            digest=hashlib.sha256(to_sign).digest(),
            signature=signature_bytes,
        )
        if not result:
            # The previous implementation caught a cryptography.exceptions.InvalidSignature
            # exception here and re-raised the DigestSignatureError to avoid the stack trace
            # leaking any information about the key.
            raise DigestSignatureError(bucket, key)

    def _create_string_to_sign(self, digest_data, inflated_digest):
        previous_signature = digest_data['previousDigestSignature']
        if previous_signature is None:
            # The value must be 'null' to match the Java implementation.
            previous_signature = 'null'

        string_to_sign = f"{digest_data['digestEndTime']}\n{digest_data['digestS3Bucket']}/{digest_data['digestS3Object']}\n{hashlib.sha256(inflated_digest).hexdigest()}\n{previous_signature}"
        LOG.debug(f'Digest string to sign: {string_to_sign}')
        return string_to_sign.encode()


class CloudTrailValidateLogs(BasicCommand):
    """
    Validates log digests and log files, optionally saving them to disk.
    """

    NAME = 'validate-logs'
    DESCRIPTION = """
    Validates CloudTrail logs for a given period of time.

    This command uses the digest files delivered to your S3 bucket to perform
    the validation. It supports validation of both digest files and
    backfill digest files in a single run.

    The AWS CLI allows you to detect the following types of changes:

    - Modification or deletion of CloudTrail log files.
    - Modification or deletion of CloudTrail digest files.

    To validate log files with the AWS CLI, the following preconditions must
    be met:

    - You must have online connectivity to AWS.
    - You must have read access to the S3 bucket that contains the digest and
      log files.
    - The digest and log files must not have been moved from the original S3
      location where CloudTrail delivered them.
    - For organization trails you must have access to describe-organization to
      validate digest files

    When you disable Log File Validation, the chain of digest files is broken
    after one hour. CloudTrail will not digest log files that were delivered
    during a period in which the Log File Validation feature was disabled.
    For example, if you enable Log File Validation on January 1, disable it
    on January 2, and re-enable it on January 10, digest files will not be
    created for the log files delivered from January 3 to January 9. The same
    applies whenever you stop CloudTrail logging or delete a trail.

    .. note::

        Log files that have been downloaded to local disk cannot be validated
        with the AWS CLI. The CLI will download all log files each time this
        command is executed.

    .. note::

        This command requires that the role executing the command has
        permission to call ListObjects, GetObject, and GetBucketLocation for
        each bucket referenced by the trail.

    """

    ARG_TABLE = [
        {
            'name': 'trail-arn',
            'required': True,
            'cli_type_name': 'string',
            'help_text': 'Specifies the ARN of the trail to be validated',
        },
        {
            'name': 'start-time',
            'required': True,
            'cli_type_name': 'string',
            'help_text': (
                'Specifies that log files delivered on or after the '
                'specified UTC timestamp value will be validated. '
                'Example: "2015-01-08T05:21:42Z".'
            ),
        },
        {
            'name': 'end-time',
            'cli_type_name': 'string',
            'help_text': (
                'Optionally specifies that log files delivered on or '
                'before the specified UTC timestamp value will be '
                'validated. The default value is the current time. '
                'Example: "2015-01-08T12:31:41Z".'
            ),
        },
        {
            'name': 's3-bucket',
            'cli_type_name': 'string',
            'help_text': (
                'Optionally specifies the S3 bucket where the digest '
                'files are stored. If a bucket name is not specified, '
                'the CLI will retrieve it by calling describe_trails'
            ),
        },
        {
            'name': 's3-prefix',
            'cli_type_name': 'string',
            'help_text': (
                'Optionally specifies the optional S3 prefix where the '
                'digest files are stored. If not specified, the CLI '
                'will determine the prefix automatically by calling '
                'describe_trails.'
            ),
        },
        {
            'name': 'account-id',
            'cli_type_name': 'string',
            'help_text': (
                'Optionally specifies the account for validating logs. '
                'This parameter is needed for organization trails '
                'for validating logs for specific account inside an '
                'organization'
            ),
        },
        {
            'name': 'verbose',
            'cli_type_name': 'boolean',
            'action': 'store_true',
            'help_text': 'Display verbose log validation information',
        },
    ]

    def __init__(self, session):
        super().__init__(session)
        self.trail_arn = None
        self.is_verbose = False
        self.start_time = None
        self.end_time = None
        self.s3_bucket = None
        self.s3_prefix = None
        self.s3_client_provider = None
        self.cloudtrail_client = None
        self.account_id = None
        self._source_region = None
        self._valid_digests = 0
        self._invalid_digests = 0
        self._valid_backfill_digests = 0
        self._invalid_backfill_digests = 0
        self._valid_logs = 0
        self._invalid_logs = 0
        self._is_last_status_double_space = True
        self._found_start_time = None
        self._found_end_time = None

    def _run_main(self, args, parsed_globals):
        self.handle_args(args)
        self.setup_services(parsed_globals)
        self._call()
        total_invalid_digests = (
            self._invalid_digests + self._invalid_backfill_digests
        )
        if total_invalid_digests > 0 or self._invalid_logs > 0:
            return 1
        return 0

    def handle_args(self, args):
        self.trail_arn = args.trail_arn
        self.is_verbose = args.verbose
        self.s3_bucket = args.s3_bucket
        self.s3_prefix = args.s3_prefix
        self.account_id = args.account_id
        self.start_time = normalize_date(parse_date(args.start_time))
        if args.end_time:
            self.end_time = normalize_date(parse_date(args.end_time))
        else:
            self.end_time = normalize_date(datetime.utcnow())
        if self.start_time > self.end_time:
            raise ValueError(
                'Invalid time range specified: start-time must '
                'occur before end-time'
            )

    def setup_services(self, parsed_globals):
        self._source_region = parsed_globals.region
        # Use the the same region as the region of the CLI to get locations.
        self.s3_client_provider = S3ClientProvider(
            self._session, self._source_region
        )
        client_args = {
            'region_name': parsed_globals.region,
            'verify': parsed_globals.verify_ssl,
        }
        self.organization_client = self._session.create_client(
            'organizations', **client_args
        )

        if parsed_globals.endpoint_url is not None:
            client_args['endpoint_url'] = parsed_globals.endpoint_url
        self.cloudtrail_client = self._session.create_client(
            'cloudtrail', **client_args
        )

    def _call(self):
        traverser = create_digest_traverser(
            trail_arn=self.trail_arn,
            cloudtrail_client=self.cloudtrail_client,
            organization_client=self.organization_client,
            trail_source_region=self._source_region,
            s3_client_provider=self.s3_client_provider,
            bucket=self.s3_bucket,
            prefix=self.s3_prefix,
            on_missing=self._on_missing_digest,
            on_invalid=self._on_invalid_digest,
            on_gap=self._on_digest_gap,
            account_id=self.account_id,
        )
        self._write_startup_text()

        digests = traverser.traverse_digests(
            self.start_time, self.end_time, is_backfill=False
        )
        for digest in digests:
            # Only valid digests are yielded and only valid digests can adjust
            # the found times that are reported in the CLI output summary.
            self._track_found_times(digest)

            self._valid_digests += 1

            self._write_status(
                f'Digest file\ts3://{digest["digestS3Bucket"]}/{digest["digestS3Object"]}\tvalid'
            )

            if not digest['logFiles']:
                continue
            for log in digest['logFiles']:
                self._download_log(log)

        backfill_digests = traverser.traverse_digests(
            self.start_time, self.end_time, is_backfill=True
        )
        for digest in backfill_digests:
            # Only valid digests are yielded and only valid digests can adjust
            # the found times that are reported in the CLI output summary.
            self._track_found_times(digest)

            self._valid_backfill_digests += 1

            self._write_status(
                f'(backfill) Digest file\ts3://{digest["digestS3Bucket"]}/{digest["digestS3Object"]}\tvalid'
            )

            if not digest['logFiles']:
                continue
            for log in digest['logFiles']:
                self._download_log(log)

        self._write_summary_text()

    def _track_found_times(self, digest):
        # Track the earliest found start time, but do not use a date before
        # the user supplied start date.
        digest_start_time = parse_date(digest['digestStartTime'])
        earliest_start_time = max(digest_start_time, self.start_time)
        if (
            not self._found_start_time
            or earliest_start_time < self._found_start_time
        ):
            self._found_start_time = earliest_start_time
        # Track the latest found end time from all digest types, but do not exceed
        # the user supplied end time (or the current date).
        digest_end_time = parse_date(digest['digestEndTime'])
        latest_end_time = min(digest_end_time, self.end_time)
        if not self._found_end_time or latest_end_time > self._found_end_time:
            self._found_end_time = latest_end_time

    def _download_log(self, log):
        """Download a log, decompress, and compare SHA256 checksums"""
        try:
            # Create a client that can work with this bucket.
            client = self.s3_client_provider.get_client(log['s3Bucket'])
            response = client.get_object(
                Bucket=log['s3Bucket'], Key=log['s3Object']
            )
            gzip_inflater = zlib.decompressobj(zlib.MAX_WBITS | 16)
            rolling_hash = hashlib.sha256()
            for chunk in iter(lambda: response['Body'].read(2048), b""):
                data = gzip_inflater.decompress(chunk)
                rolling_hash.update(data)
            remaining_data = gzip_inflater.flush()
            if remaining_data:
                rolling_hash.update(remaining_data)
            computed_hash = rolling_hash.hexdigest()
            if computed_hash != log['hashValue']:
                self._on_log_invalid(log)
            else:
                self._valid_logs += 1
                self._write_status(
                    f'Log file\ts3://{log["s3Bucket"]}/{log["s3Object"]}\tvalid'
                )
        except ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchKey':
                raise
            self._on_missing_log(log)
        except Exception:
            self._on_invalid_log_format(log)

    def _write_status(self, message, is_error=False):
        if is_error:
            if self._is_last_status_double_space:
                sys.stderr.write(f"{message}\n\n")
            else:
                sys.stderr.write(f"\n{message}\n\n")
            self._is_last_status_double_space = True
        elif self.is_verbose:
            self._is_last_status_double_space = False
            sys.stdout.write(f"{message}\n")

    def _write_startup_text(self):
        sys.stdout.write(
            f'Validating log files for trail {self.trail_arn} between {format_display_date(self.start_time)} and {format_display_date(self.end_time)}\n\n'
        )

    def _write_summary_text(self):
        if not self._is_last_status_double_space:
            sys.stdout.write('\n')
        sys.stdout.write(
            f'Results requested for {format_display_date(self.start_time)} to {format_display_date(self.end_time)}\n'
        )

        total_valid_digests = (
            self._valid_digests + self._valid_backfill_digests
        )
        total_invalid_digests = (
            self._invalid_digests + self._invalid_backfill_digests
        )

        if not total_valid_digests and not total_invalid_digests:
            sys.stdout.write('No digests found\n')
            return
        if not self._found_start_time or not self._found_end_time:
            sys.stdout.write('No valid digests found in range\n')
        else:
            sys.stdout.write(
                f'Results found for {format_display_date(self._found_start_time)} to {format_display_date(self._found_end_time)}:\n'
            )

        self._write_ratio(self._valid_digests, self._invalid_digests, 'digest')
        self._write_ratio(
            self._valid_backfill_digests,
            self._invalid_backfill_digests,
            'backfill digest',
        )
        self._write_ratio(self._valid_logs, self._invalid_logs, 'log')

        sys.stdout.write('\n')

    def _write_ratio(self, valid, invalid, name):
        total = valid + invalid
        if total > 0:
            sys.stdout.write(f'\n{valid}/{total} {name} files valid')
            if invalid > 0:
                sys.stdout.write(f', {invalid}/{total} {name} files INVALID')

    def _on_missing_digest(
        self, bucket, last_key, is_backfill=False, **kwargs
    ):
        if is_backfill:
            self._invalid_backfill_digests += 1
        else:
            self._invalid_digests += 1
        digest_type = '(backfill) ' if is_backfill else ''
        self._write_status(
            f'{digest_type}Digest file\ts3://{bucket}/{last_key}\tINVALID: not found',
            True,
        )

    def _on_digest_gap(self, is_backfill=False, **kwargs):
        log_type = '(backfill) ' if is_backfill else ''
        self._write_status(
            f'{log_type}No log files were delivered by CloudTrail between {format_display_date(kwargs["next_end_date"])} and {format_display_date(kwargs["last_start_date"])}',
            True,
        )

    def _on_invalid_digest(self, message, is_backfill=False, **kwargs):
        if is_backfill:
            self._invalid_backfill_digests += 1
        else:
            self._invalid_digests += 1
        digest_type = '(backfill) ' if is_backfill else ''
        self._write_status(f'{digest_type}{message}', True)

    def _on_invalid_log_format(self, log_data):
        self._invalid_logs += 1
        self._write_status(
            f'Log file\ts3://{log_data["s3Bucket"]}/{log_data["s3Object"]}\tINVALID: invalid format',
            True,
        )

    def _on_log_invalid(self, log_data):
        self._invalid_logs += 1
        self._write_status(
            f"Log file\ts3://{log_data['s3Bucket']}/{log_data['s3Object']}\tINVALID: hash value doesn't match",
            True,
        )

    def _on_missing_log(self, log_data):
        self._invalid_logs += 1
        self._write_status(
            f'Log file\ts3://{log_data["s3Bucket"]}/{log_data["s3Object"]}\tINVALID: not found',
            True,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from botocore import xform_name
from botocore.exceptions import UnknownServiceError

import awscli.clidriver
from awscli.autocomplete.db import DatabaseConnection


def create_apicall_indexer(filename):
    index = APICallIndexer(DatabaseConnection(filename))
    return index


class APICallIndexer:
    _CREATE_APICALL_TABLE = """\
        CREATE TABLE IF NOT EXISTS apicall_table (
          apicall_data TEXT,
          param_id INTEGER UNIQUE,
          FOREIGN KEY (param_id) REFERENCES
            param_table(param_id)
        );
    """
    _INSERT_COMMAND = """\
        INSERT OR REPLACE INTO apicall_table (apicall_data, param_id)
        VALUES
          (:apicall_data,
           (SELECT param_id FROM param_table
            WHERE
              argname = :argname AND
              command = :command AND
              parent = :parent))
    """

    def __init__(self, db_connection):
        self._db_connection = db_connection

    def generate_index(self, clidriver):
        self._create_tables()
        session = clidriver.session
        loader = session.get_component('data_loader')
        for key, command in self._iter_all_commands(clidriver):
            self._construct_completion_data(loader, command)

    def _iter_all_commands(self, clidriver):
        stack = sorted(clidriver.subcommand_table.items())
        while stack:
            key, command = stack.pop()
            for subkey, subcommand in sorted(command.subcommand_table.items()):
                stack.append((subkey, subcommand))
            yield key, command

    def _create_tables(self):
        self._db_connection.execute(self._CREATE_APICALL_TABLE)

    def _construct_completion_data(self, loader, command):
        # This method uses internal attributes to decided whether or not
        # to generate server side completion data.
        # This is using internal attributes and specific command subclasses
        # to retrieve the needed completion info.  There's no public
        # way to access this information.
        if not isinstance(command, awscli.clidriver.ServiceOperation):
            return
        # First, if there's no completions-1.json file, we can immediately
        # return.
        op_model = command._operation_model
        service_name = op_model.service_model.service_name
        try:
            completions = loader.load_service_model(
                service_name, type_name='completions-1'
            )
        except UnknownServiceError:
            return None
        # The completions-1 file is for the entire service.  We need
        # to now check if there's completion data for this specific
        # operation.
        if op_model.name not in completions['operations']:
            return
        completion_for_op = completions['operations'][op_model.name]
        for arg_name, arg_obj in sorted(command.arg_table.items()):
            if not hasattr(arg_obj, '_serialized_name'):
                continue
            api_casing = arg_obj._serialized_name
            if api_casing not in completion_for_op:
                continue
            # At this point we know there's completion info we need.
            transformed = self._transform_completion_data(
                completion_for_op[api_casing],
                completions['resources'],
                service_name,
            )
            self._insert_into_db(transformed, arg_name, command)

    def _insert_into_db(self, transformed, arg_name, command):
        parent = '.'.join(['aws'] + command.lineage_names[:-1])
        self._db_connection.execute(
            self._INSERT_COMMAND,
            apicall_data=json.dumps(transformed),
            argname=arg_name,
            command=command.name,
            parent=parent,
        )

    def _transform_completion_data(
        self, completions_for_op, resources, service_name
    ):
        # The completions-1.json data is strictly model based.  That is,
        # it's based entirely on the service API and has no mention of CLI
        # commands, parameters, etc.  This method attempts to map
        # the CLI specific commands parameters to the data in
        # completions-1.json.
        # The final format we need is:
        #
        # {'completions': [{
        #    'parameters': {},
        #    'service': 'client-name',
        #    'operation': 'snake_case_method',
        #    'jp_expr': 'jmespath[].expr',
        #  }]
        # }
        #
        # This transformation minimizes the amount of processing needed
        # by the completer during completion time.
        completions = []
        for completion in completions_for_op['completions']:
            resource = resources[completion['resourceName']]
            jp_expr = resource['resourceIdentifier'][
                completion['resourceIdentifier']
            ]
            transformed = {
                'parameters': completion['parameters'],
                'service': service_name,
                'operation': xform_name(resource['operation']),
                'jp_expr': jp_expr,
            }
            completions.append(transformed)
        return {'completions': completions}



import logging
import os

from botocore.credentials import JSONFileCache
from botocore.exceptions import ProfileNotFound

LOG = logging.getLogger(__name__)
CACHE_DIR = os.path.expanduser(os.path.join('~', '.aws', 'cli', 'cache'))


def register_assume_role_provider(event_handlers):
    event_handlers.register(
        'session-initialized',
        inject_assume_role_provider_cache,
        unique_id='inject_assume_role_cred_provider_cache',
    )


def inject_assume_role_provider_cache(session, **kwargs):
    try:
        cred_chain = session.get_component('credential_provider')
    except ProfileNotFound:
        # If a user has provided a profile that does not exist,
        # trying to retrieve components/config on the session
        # will raise ProfileNotFound.  Sometimes this is invalid:
        #
        # "ec2 describe-instances --profile unknown"
        #
        # and sometimes this is perfectly valid:
        #
        # "configure set region us-west-2 --profile brand-new-profile"
        #
        # Because we can't know (and don't want to know) whether
        # the customer is trying to do something valid, we just
        # immediately return.  If it's invalid something else
        # up the stack will raise ProfileNotFound, otherwise
        # the configure (and other) commands will work as expected.
        LOG.debug(
            "ProfileNotFound caught when trying to inject "
            "assume-role cred provider cache.  Not configuring "
            "JSONFileCache for assume-role."
        )
        return
    assume_role_provider = cred_chain.get_provider('assume-role')
    assume_role_provider.cache = JSONFileCache(CACHE_DIR)
    web_identity_provider = cred_chain.get_provider(
        'assume-role-with-web-identity'
    )
    web_identity_provider.cache = JSONFileCache(CACHE_DIR)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand, BasicHelp
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.wizard import devcommands, factory
from awscli.customizations.wizard.loader import WizardLoader


def register_wizard_commands(event_handlers):
    devcommands.register_dev_commands(event_handlers)
    loader = WizardLoader()
    commands = loader.list_commands_with_wizards()
    _register_wizards_for_commands(commands, event_handlers)


def _register_wizards_for_commands(commands, event_handlers):
    for command in commands:
        event_handlers.register(
            'building-command-table.%s' % command, _add_wizard_command
        )


def _add_wizard_command(session, command_object, command_table, **kwargs):
    v1_runner = factory.create_default_wizard_v1_runner(session)
    v2_runner = factory.create_default_wizard_v2_runner(session)
    cmd = TopLevelWizardCommand(
        session=session,
        loader=WizardLoader(),
        parent_command=command_object.name,
        runner={'0.1': v1_runner, '0.2': v2_runner},
    )
    command_table['wizard'] = cmd


class TopLevelWizardCommand(BasicCommand):
    NAME = 'wizard'
    DESCRIPTION = (
        'Interactive command for creating and configuring AWS resources.'
    )

    def __init__(
        self, session, loader, parent_command, runner, wizard_name='_main'
    ):
        super(TopLevelWizardCommand, self).__init__(session)
        self._session = session
        self._loader = loader
        self._parent_command = parent_command
        self._runner = runner
        self._wizard_name = wizard_name

    def _build_subcommand_table(self):
        subcommand_table = super(
            TopLevelWizardCommand, self
        )._build_subcommand_table()
        wizards = self._get_available_wizards()
        for name in wizards:
            cmd = SingleWizardCommand(
                self._session,
                self._loader,
                self._parent_command,
                self._runner,
                wizard_name=name,
            )
            subcommand_table[name] = cmd
        self._add_lineage(subcommand_table)
        return subcommand_table

    def _get_available_wizards(self):
        wizards = self._loader.list_available_wizards(self._parent_command)
        return [name for name in wizards if not name.startswith('_')]

    def _run_main(self, parsed_args, parsed_globals):
        if self._wizard_exists():
            self._run_wizard()
            return 0
        else:
            self._raise_usage_error()

    def _wizard_exists(self):
        return self._loader.wizard_exists(
            self._parent_command, self._wizard_name
        )

    def _run_wizard(self):
        loaded = self._loader.load_wizard(
            self._parent_command, self._wizard_name
        )
        version = loaded.get('version')
        if version in self._runner:
            self._runner[version].run(loaded)
        else:
            raise ParamValidationError(
                'Definition file has unsupported version %s ' % version
            )

    def create_help_command(self):
        return BasicHelp(
            self._session,
            self,
            command_table=self.subcommand_table,
            arg_table=self.arg_table,
        )


class SingleWizardCommand(TopLevelWizardCommand):
    def __init__(self, session, loader, parent_command, runner, wizard_name):
        super(SingleWizardCommand, self).__init__(
            session, loader, parent_command, runner, wizard_name
        )
        self._session = session
        self._loader = loader
        self._runner = runner
        self._wizard_name = wizard_name
        self.NAME = self._wizard_name

    def _build_subcommand_table(self):
        return {}

    def _run_main(self, parsed_args, parsed_globals):
        self._run_wizard()
        return 0

    def create_help_command(self):
        loaded = self._loader.load_wizard(
            self._parent_command,
            self._wizard_name,
        )
        return WizardHelpCommand(
            self._session, self, self.subcommand_table, self.arg_table, loaded
        )


class WizardHelpCommand(BasicHelp):
    def __init__(
        self, session, command_object, command_table, arg_table, loaded_wizard
    ):
        super(WizardHelpCommand, self).__init__(
            session, command_object, command_table, arg_table
        )
        self._description = loaded_wizard.get('description', '')



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import json
import logging
import os
import socket
import time
import webbrowser
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer

from botocore.compat import parse_qs, urlparse
from botocore.credentials import JSONFileCache
from botocore.exceptions import (
    AuthCodeFetcherError,
    PendingAuthorizationExpiredError,
)
from botocore.utils import (
    SSOTokenFetcher,
    SSOTokenFetcherAuth,
    original_ld_library_path,
)

from awscli import __version__ as awscli_version
from awscli.customizations.assumerole import CACHE_DIR as AWS_CREDS_CACHE_DIR
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ConfigurationError
from awscli.customizations.utils import uni_print

LOG = logging.getLogger(__name__)

SSO_TOKEN_DIR = os.path.expanduser(os.path.join('~', '.aws', 'sso', 'cache'))

LOGIN_ARGS = [
    {
        'name': 'no-browser',
        'action': 'store_true',
        'default': False,
        'help_text': (
            'Disables automatically opening the verification URL in the '
            'default browser.'
        ),
    },
    {
        'name': 'use-device-code',
        'action': 'store_true',
        'default': False,
        'help_text': (
            'Uses the Device Code authorization grant and login flow '
            'instead of the Authorization Code flow.'
        ),
    },
]


def _serialize_utc_timestamp(obj):
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
    return obj


def _sso_json_dumps(obj):
    return json.dumps(obj, default=_serialize_utc_timestamp)


def do_sso_login(
    session,
    sso_region,
    start_url,
    parsed_globals,
    token_cache=None,
    on_pending_authorization=None,
    force_refresh=False,
    registration_scopes=None,
    session_name=None,
    use_device_code=False,
):
    if token_cache is None:
        token_cache = JSONFileCache(SSO_TOKEN_DIR, dumps_func=_sso_json_dumps)
    if on_pending_authorization is None:
        on_pending_authorization = OpenBrowserHandler(
            open_browser=open_browser_with_original_ld_path
        )

    # For the auth flow, we need a non-legacy sso-session and check that the
    # user hasn't opted into falling back to the device code flow
    if session_name and not use_device_code:
        token_fetcher = SSOTokenFetcherAuth(
            sso_region=sso_region,
            client_creator=session.create_client,
            parsed_globals=parsed_globals,
            auth_code_fetcher=AuthCodeFetcher(),
            cache=token_cache,
            on_pending_authorization=on_pending_authorization,
        )
    else:
        token_fetcher = SSOTokenFetcher(
            sso_region=sso_region,
            client_creator=session.create_client,
            parsed_globals=parsed_globals,
            cache=token_cache,
            on_pending_authorization=on_pending_authorization,
        )

    return token_fetcher.fetch_token(
        start_url=start_url,
        session_name=session_name,
        force_refresh=force_refresh,
        registration_scopes=registration_scopes,
    )


def parse_sso_registration_scopes(raw_scopes):
    parsed_scopes = []
    for scope in raw_scopes.split(','):
        if scope := scope.strip():
            parsed_scopes.append(scope)
    return parsed_scopes


def open_browser_with_original_ld_path(url):
    with original_ld_library_path():
        webbrowser.open_new_tab(url)


class BaseAuthorizationhandler:
    def __call__(
        self,
        userCode,
        verificationUri,
        verificationUriComplete,
        cross_device_flag,
        **kwargs,
    ):
        # Pending authorization handlers should always take **kwargs in case
        # the API begins to return new values.
        raise NotImplementedError("authorization_handler")


class PrintOnlyHandler(BaseAuthorizationhandler):
    def __init__(self, outfile=None):
        self._outfile = outfile

    def __call__(
        self, userCode, verificationUri, verificationUriComplete, **kwargs
    ):
        opening_msg = (
            f'Browser will not be automatically opened.\n'
            f'Please visit the following URL:\n'
            f'\n{verificationUri}\n'
        )

        user_code_msg = (
            f'\nThen enter the code:\n'
            f'\n{userCode}\n'
            f'\nAlternatively, you may visit the following URL which will '
            f'autofill the code upon loading:'
            f'\n{verificationUriComplete}\n'
        )

        uni_print(opening_msg, self._outfile)
        if userCode:
            uni_print(user_code_msg, self._outfile)


class OpenBrowserHandler(BaseAuthorizationhandler):
    def __init__(self, outfile=None, open_browser=None):
        self._outfile = outfile
        if open_browser is None:
            open_browser = webbrowser.open_new_tab
        self._open_browser = open_browser

    def __call__(
        self,
        userCode,
        verificationUri,
        verificationUriComplete,
        cross_device_flag,
        **kwargs,
    ):
        if userCode:  # only the device code flow supports different devices
            opening_msg = (
                f'Attempting to open your default browser.\n'
                f'If the browser does not open or '
                f'you wish to use a different device to authorize this '
                f'request, open the following URL:\n'
                f'\n{verificationUri}\n'
            )
        else:
            opening_msg = (
                f'Attempting to open your default browser. '
                f'If the browser does not open, open the following URL.\n'
                f'If you are unable to open the URL on this device, '
                f'run this command again with the \'{cross_device_flag}\' option.\n'
                f'\n{verificationUri}\n\n'
            )

        user_code_msg = f'\nThen enter the code:\n\n{userCode}\n'
        uni_print(opening_msg, self._outfile)
        if userCode:
            uni_print(user_code_msg, self._outfile)

        if self._open_browser:
            try:
                return self._open_browser(verificationUriComplete)
            except Exception:
                LOG.debug('Failed to open browser:', exc_info=True)


class AuthCodeFetcher:
    """Manages the local web server that will be used
    to retrieve the authorization code from the OAuth callback
    """

    # How many seconds handle_request should wait for an incoming request
    _REQUEST_TIMEOUT = 10
    # How long we wait overall for the callback
    _OVERALL_TIMEOUT = 60 * 10

    def __init__(self):
        self._auth_code = None
        self._state = None
        self._is_done = False

        # We do this so that the request handler can have a reference to this
        # AuthCodeFetcher so that it can pass back the state and auth code
        try:
            handler = partial(OAuthCallbackHandler, self)
            self.http_server = HTTPServer(('', 0), handler)
            self.http_server.timeout = self._REQUEST_TIMEOUT
        except OSError as e:
            raise AuthCodeFetcherError(error_msg=e)

    def redirect_uri_without_port(self):
        return 'http://127.0.0.1/oauth/callback'

    def redirect_uri_with_port(self):
        return (
            f'http://127.0.0.1:{self.http_server.server_port}/oauth/callback'
        )

    def get_auth_code_and_state(self):
        """Blocks until the expected redirect request with either the
        authorization code/state or and error is handled
        """
        LOG.debug(f'Waiting for auth code at {self.redirect_uri_with_port()}')
        start = time.time()
        while (
            not self._is_done and time.time() < start + self._OVERALL_TIMEOUT
        ):
            self.http_server.handle_request()
        self.http_server.server_close()

        if not self._is_done:
            raise PendingAuthorizationExpiredError

        return self._auth_code, self._state

    def set_auth_code_and_state(self, auth_code, state):
        self._auth_code = auth_code
        self._state = state
        self._is_done = True


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    """HTTP handler to handle OAuth callback requests, extracting
    the auth code and state parameters, and displaying a page directing
    the user to return to the CLI.
    """

    def __init__(self, auth_code_fetcher, *args, **kwargs):
        self._auth_code_fetcher = auth_code_fetcher
        super().__init__(*args, **kwargs)

    def log_message(self, format, *args):
        # Suppress built-in logging, otherwise it prints
        # each request to console
        pass

    def version_string(self):
        # Override the Host header in case helpful for debugging
        return f'AWS CLI/{awscli_version}'

    def do_GET(self):
        self.send_response(200)
        self.end_headers()
        with open(
            os.path.join(os.path.dirname(__file__), 'index.html'),
            'rb',
        ) as file:
            self.wfile.write(file.read())

        query_params = parse_qs(urlparse(self.path).query)

        if 'error' in query_params:
            self._auth_code_fetcher.set_auth_code_and_state(
                None,
                None,
            )
        elif 'code' in query_params and 'state' in query_params:
            self._auth_code_fetcher.set_auth_code_and_state(
                query_params['code'][0],
                query_params['state'][0],
            )


class InvalidSSOConfigError(ConfigurationError):
    pass


class BaseSSOCommand(BasicCommand):
    _REQUIRED_SSO_CONFIG_VARS = [
        'sso_start_url',
        'sso_region',
    ]

    def _get_sso_config(self, sso_session=None):
        scoped_config = self._session.get_scoped_config()
        if sso_session is None:
            sso_session = scoped_config.get('sso_session')
        if sso_session:
            return self._get_sso_session_config(sso_session)
        else:
            return self._get_legacy_sso_config(scoped_config)

    def _get_sso_session_config(self, session_name):
        full_config = self._session.full_config
        if session_name not in full_config.get('sso_sessions', {}):
            raise InvalidSSOConfigError(
                f'The specified sso-session does not exist: "{session_name}"'
            )
        session_config = full_config['sso_sessions'][session_name]
        sso_config, missing = self._get_required_config_vars(session_config)
        sso_config['session_name'] = session_name

        scopes_var = 'sso_registration_scopes'
        if scopes_var in session_config:
            raw_scopes = session_config[scopes_var]
            parsed_scopes = parse_sso_registration_scopes(raw_scopes)
            sso_config['registration_scopes'] = parsed_scopes

        if missing:
            error_msg = f'Missing the following required SSO configuration values: {", ".join(missing)}. '
            raise InvalidSSOConfigError(error_msg)

        return sso_config

    def _get_legacy_sso_config(self, scoped_config):
        sso_config, missing = self._get_required_config_vars(scoped_config)
        if missing:
            raise InvalidSSOConfigError(
                f'Missing the following required SSO configuration values: {", ".join(missing)}. '
                'To make sure this profile is properly configured to use SSO, '
                'please run: aws configure sso'
            )
        return sso_config

    def _get_required_config_vars(self, config):
        sso_config = {}
        missing_vars = []
        for config_var in self._REQUIRED_SSO_CONFIG_VARS:
            if config_var not in config:
                missing_vars.append(config_var)
            else:
                sso_config[config_var] = config[config_var]
        return sso_config, missing_vars



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import re

from botocore import xform_name
from botocore.model import StringShape
from botocore.utils import is_json_value_header

from awscli import SCALAR_TYPES, __version__ as AWS_CLI_VERSION
from awscli.argprocess import ParamShorthandDocGen
from awscli.bcdoc.docevents import DOC_EVENTS
from awscli.topictags import TopicTagDB
from awscli.utils import (
    find_service_and_method_in_event_name,
    is_document_type,
    is_streaming_blob_type,
    is_tagged_union_type,
    operation_uses_document_types,
)

LOG = logging.getLogger(__name__)
EXAMPLES_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'examples'
)
GLOBAL_OPTIONS_FILE = os.path.join(EXAMPLES_DIR, 'global_options.rst')
GLOBAL_OPTIONS_SYNOPSIS_FILE = os.path.join(
    EXAMPLES_DIR, 'global_synopsis.rst'
)


class CLIDocumentEventHandler:
    def __init__(self, help_command):
        self.help_command = help_command
        self.register(help_command.session, help_command.event_class)
        self._arg_groups = self._build_arg_table_groups(help_command)
        self._documented_arg_groups = []

    def _build_arg_table_groups(self, help_command):
        arg_groups = {}
        for arg in help_command.arg_table.values():
            if arg.group_name is not None:
                arg_groups.setdefault(arg.group_name, []).append(arg)
        return arg_groups

    def _get_argument_type_name(self, shape, default):
        if is_json_value_header(shape):
            return 'JSON'
        if is_document_type(shape):
            return 'document'
        if is_streaming_blob_type(shape):
            return 'streaming blob'
        if is_tagged_union_type(shape):
            return 'tagged union structure'
        return default

    def _map_handlers(self, session, event_class, mapfn):
        for event in DOC_EVENTS:
            event_handler_name = event.replace('-', '_')
            if hasattr(self, event_handler_name):
                event_handler = getattr(self, event_handler_name)
                format_string = DOC_EVENTS[event]
                num_args = len(format_string.split('.')) - 2
                format_args = (event_class,) + ('*',) * num_args
                event_string = event + format_string % format_args
                unique_id = event_class + event_handler_name
                mapfn(event_string, event_handler, unique_id)

    def register(self, session, event_class):
        """
        The default register iterates through all of the
        available document events and looks for a corresponding
        handler method defined in the object.  If it's there, that
        handler method will be registered for the all events of
        that type for the specified ``event_class``.
        """
        self._map_handlers(session, event_class, session.register)

    def unregister(self):
        """
        The default unregister iterates through all of the
        available document events and looks for a corresponding
        handler method defined in the object.  If it's there, that
        handler method will be unregistered for the all events of
        that type for the specified ``event_class``.
        """
        self._map_handlers(
            self.help_command.session,
            self.help_command.event_class,
            self.help_command.session.unregister,
        )

    # These are default doc handlers that apply in the general case.

    def doc_breadcrumbs(self, help_command, **kwargs):
        doc = help_command.doc
        if doc.target != 'man':
            cmd_names = help_command.event_class.split('.')
            doc.write('[ ')
            doc.write(':ref:`aws <cli:aws>`')
            full_cmd_list = ['aws']
            for cmd in cmd_names[:-1]:
                doc.write(' . ')
                full_cmd_list.append(cmd)
                full_cmd_name = ' '.join(full_cmd_list)
                doc.write(f':ref:`{cmd} <cli:{full_cmd_name}>`')
            doc.writeln(' ]')
            doc.writeln('')

    def doc_title(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.new_paragraph()
        reference = help_command.event_class.replace('.', ' ')
        if reference != 'aws':
            reference = 'aws ' + reference
        doc.writeln(f'.. _cli:{reference}:')
        doc.style.h1(help_command.name)

    def doc_description(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Description')
        doc.include_doc_string(help_command.description)
        doc.style.new_paragraph()

    def doc_synopsis_start(self, help_command, **kwargs):
        self._documented_arg_groups = []
        doc = help_command.doc
        doc.style.h2('Synopsis')
        doc.style.start_codeblock()
        doc.writeln(f'{help_command.name}')

    def doc_synopsis_option(self, arg_name, help_command, **kwargs):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
            option_str = ' | '.join(
                [a.cli_name for a in self._arg_groups[argument.group_name]]
            )
            self._documented_arg_groups.append(argument.group_name)
        elif argument.cli_name.startswith('--'):
            option_str = f'{argument.cli_name} <value>'
        else:
            option_str = f'<{argument.cli_name}>'
        if not (
            argument.required
            or getattr(argument, '_DOCUMENT_AS_REQUIRED', False)
        ):
            option_str = f'[{option_str}]'
        doc.writeln(f'{option_str}')

    def doc_synopsis_end(self, help_command, **kwargs):
        doc = help_command.doc
        # Append synopsis for global options.
        doc.write_from_file(GLOBAL_OPTIONS_SYNOPSIS_FILE)
        doc.style.end_codeblock()
        # Reset the documented arg groups for other sections
        # that may document args (the detailed docs following
        # the synopsis).
        self._documented_arg_groups = []

    def doc_options_start(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Options')
        if not help_command.arg_table:
            doc.write('*None*\n')

    def doc_option(self, arg_name, help_command, **kwargs):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        if argument.required or getattr(
            argument, '_DOCUMENT_AS_REQUIRED', False
        ):
            required_field = ' [required]'
        else:
            required_field = ''

        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
            name = ' | '.join(
                [
                    f'``{a.cli_name}``'
                    for a in self._arg_groups[argument.group_name]
                ]
            )
            self._documented_arg_groups.append(argument.group_name)
        else:
            name = f'``{argument.cli_name}``'
        doc.write(
            f'{name} '
            f'({self._get_argument_type_name(argument.argument_model, argument.cli_type_name)})'
            f'{required_field}\n'
        )

        doc.style.indent()
        doc.include_doc_string(argument.documentation)
        if is_streaming_blob_type(argument.argument_model):
            self._add_streaming_blob_note(doc)
        if is_tagged_union_type(argument.argument_model):
            self._add_tagged_union_note(argument.argument_model, doc)
        if hasattr(argument, 'argument_model'):
            self._document_enums(argument.argument_model, doc)
            self._document_constraints(argument.argument_model, doc)
            self._document_nested_structure(argument.argument_model, doc)
        doc.style.dedent()
        doc.style.new_paragraph()

    def doc_global_option(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Global Options')
        doc.write_from_file(GLOBAL_OPTIONS_FILE)

    def doc_relateditems_start(self, help_command, **kwargs):
        if help_command.related_items:
            doc = help_command.doc
            doc.style.h2('See Also')

    def doc_relateditem(self, help_command, related_item, **kwargs):
        doc = help_command.doc
        doc.write('* ')
        doc.style.sphinx_reference_label(
            label=f'cli:{related_item}', text=related_item
        )
        doc.write('\n')

    def doc_meta_description(self, help_command, **kwargs):
        pass

    def _document_enums(self, model, doc):
        """Documents top-level parameter enums"""
        if isinstance(model, StringShape):
            if model.enum:
                doc.style.new_paragraph()
                doc.write('Possible values:')
                doc.style.start_ul()
                for enum in model.enum:
                    doc.style.li(f'``{enum}``')
                doc.style.end_ul()

    def _document_nested_structure(self, model, doc):
        """Recursively documents parameters in nested structures"""
        member_type_name = getattr(model, 'type_name', None)
        if member_type_name == 'structure':
            required_members = model.metadata.get('required', [])
            for member_name, member_shape in model.members.items():
                is_required = member_name in required_members
                self._doc_member(
                    doc,
                    member_name,
                    member_shape,
                    stack=[model.name],
                    required=is_required,
                )
        elif member_type_name == 'list':
            self._doc_member(
                doc, '', model.member, stack=[model.name], required=False
            )
        elif member_type_name == 'map':
            key_shape = model.key
            key_name = key_shape.serialization.get('name', 'key')
            self._doc_member(
                doc, key_name, key_shape, stack=[model.name], required=False
            )
            value_shape = model.value
            value_name = value_shape.serialization.get('name', 'value')
            self._doc_member(
                doc,
                value_name,
                value_shape,
                stack=[model.name],
                required=False,
            )

    def _doc_member(
        self, doc, member_name, member_shape, stack, required=False
    ):
        if member_shape.name in stack:
            # Document the recursion once, otherwise just
            # note the fact that it's recursive and return.
            if stack.count(member_shape.name) > 1:
                if member_shape.type_name == 'structure':
                    doc.write('( ... recursive ... )')
                return
        stack.append(member_shape.name)
        try:
            self._do_doc_member(
                doc, member_name, member_shape, stack, required
            )
        finally:
            stack.pop()

    def _do_doc_member(
        self, doc, member_name, member_shape, stack, required=False
    ):
        docs = member_shape.documentation
        type_name = self._get_argument_type_name(
            member_shape, member_shape.type_name
        )

        if member_name:
            parameter_string = f'{member_name} -> ({type_name})'
        else:
            parameter_string = f'({type_name})'

        if required:
            parameter_string += ' [required]'

        doc.write(parameter_string)

        doc.style.indent()
        doc.style.new_paragraph()
        doc.include_doc_string(docs)
        if is_tagged_union_type(member_shape):
            self._add_tagged_union_note(member_shape, doc)

        self._document_enums(member_shape, doc)
        self._document_constraints(member_shape, doc)
        doc.style.new_paragraph()
        member_type_name = member_shape.type_name
        if member_type_name == 'structure':
            required_members = member_shape.metadata.get('required', [])
            for sub_name, sub_shape in member_shape.members.items():
                sub_required = sub_name in required_members
                self._doc_member(
                    doc, sub_name, sub_shape, stack, required=sub_required
                )
        elif member_type_name == 'map':
            key_shape = member_shape.key
            key_name = key_shape.serialization.get('name', 'key')
            self._doc_member(doc, key_name, key_shape, stack, required=False)
            value_shape = member_shape.value
            value_name = value_shape.serialization.get('name', 'value')
            self._doc_member(
                doc, value_name, value_shape, stack, required=False
            )
        elif member_type_name == 'list':
            self._doc_member(
                doc, '', member_shape.member, stack, required=False
            )
        doc.style.dedent()
        doc.style.new_paragraph()

    def _add_streaming_blob_note(self, doc):
        doc.style.start_note()
        msg = (
            "This argument is of type: streaming blob. "
            "Its value must be the path to a file "
            "(e.g. ``path/to/file``) and must **not** "
            "be prefixed with ``file://`` or ``fileb://``"
        )
        doc.writeln(msg)
        doc.style.end_note()

    def _add_tagged_union_note(self, shape, doc):
        doc.style.start_note()
        members_str = ", ".join([f'``{key}``' for key in shape.members.keys()])
        msg = (
            "This is a Tagged Union structure. Only one of the "
            f"following top level keys can be set: {members_str}."
        )
        doc.writeln(msg)
        doc.style.end_note()

    def _document_constraints(self, model, doc):
        """Documents parameter value constraints"""
        if not hasattr(model, 'metadata'):
            return
        constraints = ['min', 'max', 'pattern']
        if not any(
            [constraint in model.metadata for constraint in constraints]
        ):
            return
        doc.style.new_paragraph()
        doc.write('Constraints:')
        doc.style.start_ul()
        for constraint in constraints:
            if (val := model.metadata.get(constraint)) is not None:
                doc.style.li(f'{constraint}: ``{val}``')
        doc.style.end_ul()


class ProviderDocumentEventHandler(CLIDocumentEventHandler):
    def doc_breadcrumbs(self, help_command, event_name, **kwargs):
        pass

    def doc_synopsis_start(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Synopsis')
        doc.style.codeblock(help_command.synopsis)
        doc.include_doc_string(help_command.help_usage)

    def doc_synopsis_option(self, arg_name, help_command, **kwargs):
        pass

    def doc_synopsis_end(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.new_paragraph()

    def doc_options_start(self, help_command, **kwargs):
        pass

    def doc_option(self, arg_name, help_command, **kwargs):
        pass

    def doc_subitems_start(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Available Services')
        doc.style.toctree()

    def doc_subitem(self, command_name, help_command, **kwargs):
        doc = help_command.doc
        file_name = f'{command_name}/index'
        doc.style.tocitem(command_name, file_name=file_name)


class ServiceDocumentEventHandler(CLIDocumentEventHandler):
    # A service document has no synopsis.
    def doc_synopsis_start(self, help_command, **kwargs):
        pass

    def doc_synopsis_option(self, arg_name, help_command, **kwargs):
        pass

    def doc_synopsis_end(self, help_command, **kwargs):
        pass

    # A service document has no option section.
    def doc_options_start(self, help_command, **kwargs):
        pass

    def doc_option(self, arg_name, help_command, **kwargs):
        pass

    def doc_option_example(self, arg_name, help_command, **kwargs):
        pass

    def doc_options_end(self, help_command, **kwargs):
        pass

    def doc_global_option(self, help_command, **kwargs):
        pass

    def doc_description(self, help_command, **kwargs):
        doc = help_command.doc
        service_model = help_command.obj
        doc.style.h2('Description')
        # TODO: need a documentation attribute.
        doc.include_doc_string(service_model.documentation)

    def doc_subitems_start(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Available Commands')
        doc.style.toctree()

    def doc_subitem(self, command_name, help_command, **kwargs):
        doc = help_command.doc
        subcommand = help_command.command_table[command_name]
        subcommand_table = getattr(subcommand, 'subcommand_table', {})
        # If the subcommand table has commands in it,
        # direct the subitem to the command's index because
        # it has more subcommands to be documented.
        if len(subcommand_table) > 0:
            file_name = f'{command_name}/index'
            doc.style.tocitem(command_name, file_name=file_name)
        else:
            doc.style.tocitem(command_name)

    def doc_meta_description(self, help_command, **kwargs):
        doc = help_command.doc
        reference = help_command.event_class.replace('.', ' ')
        doc.writeln(".. meta::")
        doc.writeln(f"   :description: Learn about the AWS CLI {AWS_CLI_VERSION} {reference} commands.")
        doc.writeln("")


class OperationDocumentEventHandler(CLIDocumentEventHandler):
    AWS_DOC_BASE = 'https://docs.aws.amazon.com/goto/WebAPI'

    def doc_description(self, help_command, **kwargs):
        doc = help_command.doc
        operation_model = help_command.obj
        doc.style.h2('Description')
        doc.include_doc_string(operation_model.documentation)
        self._add_webapi_crosslink(help_command)
        self._add_note_for_document_types_if_used(help_command)

    def _add_webapi_crosslink(self, help_command):
        doc = help_command.doc
        operation_model = help_command.obj
        service_model = operation_model.service_model
        service_uid = service_model.metadata.get('uid')
        if service_uid is None:
            # If there's no service_uid in the model, we can't
            # be certain if the generated cross link will work
            # so we don't generate any crosslink info.
            return
        doc.style.new_paragraph()
        doc.write("See also: ")
        link = f'{self.AWS_DOC_BASE}/{service_uid}/{operation_model.name}'
        doc.style.external_link(title="AWS API Documentation", link=link)
        doc.writeln('')

    def _add_note_for_document_types_if_used(self, help_command):
        if operation_uses_document_types(help_command.obj):
            help_command.doc.style.new_paragraph()
            help_command.doc.writeln(
                f'``{help_command.name}`` uses document type values. Document types follow the '
                'JSON data model where valid values are: strings, numbers, '
                'booleans, null, arrays, and objects. For command input, '
                'options and nested parameters that are labeled with the type '
                '``document`` must be provided as JSON. Shorthand syntax does '
                'not support document types.'
            )

    def _json_example_value_name(
        self, argument_model, include_enum_values=True
    ):
        # If include_enum_values is True, then the valid enum values
        # are included as the sample JSON value.
        if isinstance(argument_model, StringShape):
            if argument_model.enum and include_enum_values:
                choices = argument_model.enum
                return '|'.join([f'"{c}"' for c in choices])
            else:
                return '"string"'
        elif argument_model.type_name == 'boolean':
            return 'true|false'
        else:
            return f'{argument_model.type_name}'

    def _json_example(self, doc, argument_model, stack):
        if argument_model.name in stack:
            # Document the recursion once, otherwise just
            # note the fact that it's recursive and return.
            if stack.count(argument_model.name) > 1:
                if argument_model.type_name == 'structure':
                    doc.write('{ ... recursive ... }')
                return
        stack.append(argument_model.name)
        try:
            self._do_json_example(doc, argument_model, stack)
        finally:
            stack.pop()

    def _do_json_example(self, doc, argument_model, stack):
        if argument_model.type_name == 'list':
            doc.write('[')
            if argument_model.member.type_name in SCALAR_TYPES:
                doc.write(
                    f'{self._json_example_value_name(argument_model.member)}, ...'
                )
            else:
                doc.style.indent()
                doc.style.new_line()
                self._json_example(doc, argument_model.member, stack)
                doc.style.new_line()
                doc.write('...')
                doc.style.dedent()
                doc.style.new_line()
            doc.write(']')
        elif argument_model.type_name == 'map':
            doc.write('{')
            doc.style.indent()
            key_string = self._json_example_value_name(argument_model.key)
            doc.write(f'{key_string}: ')
            if argument_model.value.type_name in SCALAR_TYPES:
                doc.write(self._json_example_value_name(argument_model.value))
            else:
                doc.style.indent()
                self._json_example(doc, argument_model.value, stack)
                doc.style.dedent()
            doc.style.new_line()
            doc.write('...')
            doc.style.dedent()
            doc.write('}')
        elif argument_model.type_name == 'structure':
            if argument_model.is_document_type:
                self._doc_document_member(doc)
            else:
                self._doc_input_structure_members(doc, argument_model, stack)

    def _doc_document_member(self, doc):
        doc.write('{...}')

    def _doc_input_structure_members(self, doc, argument_model, stack):
        doc.write('{')
        doc.style.indent()
        doc.style.new_line()
        members = argument_model.members
        for i, member_name in enumerate(members):
            member_model = members[member_name]
            member_type_name = member_model.type_name
            if member_type_name in SCALAR_TYPES:
                doc.write(
                    f'"{member_name}": {self._json_example_value_name(member_model)}'
                )
            elif member_type_name == 'structure':
                doc.write(f'"{member_name}": ')
                self._json_example(doc, member_model, stack)
            elif member_type_name == 'map':
                doc.write(f'"{member_name}": ')
                self._json_example(doc, member_model, stack)
            elif member_type_name == 'list':
                doc.write(f'"{member_name}": ')
                self._json_example(doc, member_model, stack)
            if i < len(members) - 1:
                doc.write(',')
                doc.style.new_line()
        doc.style.dedent()
        doc.style.new_line()
        doc.write('}')

    def doc_option_example(self, arg_name, help_command, event_name, **kwargs):
        service_id, operation_name = find_service_and_method_in_event_name(
            event_name
        )
        doc = help_command.doc
        cli_argument = help_command.arg_table[arg_name]
        if cli_argument.group_name in self._arg_groups:
            if cli_argument.group_name in self._documented_arg_groups:
                # Args with group_names (boolean args) don't
                # need to generate example syntax.
                return
        argument_model = cli_argument.argument_model
        docgen = ParamShorthandDocGen()
        if docgen.supports_shorthand(cli_argument.argument_model):
            example_shorthand_syntax = docgen.generate_shorthand_example(
                cli_argument, service_id, operation_name
            )
            if example_shorthand_syntax is None:
                # If the shorthand syntax returns a value of None,
                # this indicates to us that there is no example
                # needed for this param so we can immediately
                # return.
                return
            if example_shorthand_syntax:
                doc.style.new_paragraph()
                doc.write('Shorthand Syntax')
                doc.style.start_codeblock()
                for example_line in example_shorthand_syntax.splitlines():
                    doc.writeln(example_line)
                doc.style.end_codeblock()
        if (
            argument_model is not None
            and argument_model.type_name == 'list'
            and argument_model.member.type_name in SCALAR_TYPES
        ):
            # A list of scalars is special.  While you *can* use
            # JSON ( ["foo", "bar", "baz"] ), you can also just
            # use the argparse behavior of space separated lists.
            # "foo" "bar" "baz".  In fact we don't even want to
            # document the JSON syntax in this case.
            member = argument_model.member
            doc.style.new_paragraph()
            doc.write('Syntax')
            doc.style.start_codeblock()
            example_type = self._json_example_value_name(
                member, include_enum_values=False
            )
            doc.write(f'{example_type} {example_type} ...')
            doc.style.end_codeblock()
            doc.style.new_paragraph()
        elif cli_argument.cli_type_name not in SCALAR_TYPES:
            doc.style.new_paragraph()
            doc.write('JSON Syntax')
            doc.style.start_codeblock()
            self._json_example(doc, argument_model, stack=[])
            doc.style.end_codeblock()
            doc.style.new_paragraph()

    def doc_output(self, help_command, event_name, **kwargs):
        doc = help_command.doc
        doc.style.h2('Output')
        operation_model = help_command.obj
        output_shape = operation_model.output_shape
        if output_shape is None or not output_shape.members:
            doc.write('None')
        else:
            for member_name, member_shape in output_shape.members.items():
                self._doc_member(doc, member_name, member_shape, stack=[])

    def doc_meta_description(self, help_command, **kwargs):
        doc = help_command.doc
        reference = help_command.event_class.replace('.', ' ')
        doc.writeln(".. meta::")
        doc.writeln(f"   :description: Use the AWS CLI {AWS_CLI_VERSION} to run the {reference} command.")
        doc.writeln("")

class TopicListerDocumentEventHandler(CLIDocumentEventHandler):
    DESCRIPTION = (
        'This is the AWS CLI Topic Guide. It gives access to a set '
        'of topics that provide a deeper understanding of the CLI. To access '
        'the list of topics from the command line, run ``aws help topics``. '
        'To access a specific topic from the command line, run '
        '``aws help [topicname]``, where ``topicname`` is the name of the '
        'topic as it appears in the output from ``aws help topics``.'
    )

    def __init__(self, help_command):
        self.help_command = help_command
        self.register(help_command.session, help_command.event_class)
        self._topic_tag_db = TopicTagDB()
        self._topic_tag_db.load_json_index()

    def doc_breadcrumbs(self, help_command, **kwargs):
        doc = help_command.doc
        if doc.target != 'man':
            doc.write('[ ')
            doc.style.sphinx_reference_label(label='cli:aws', text='aws')
            doc.write(' ]')

    def doc_title(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.new_paragraph()
        doc.style.link_target_definition(
            refname=f'cli:aws help {self.help_command.name}', link=''
        )
        doc.style.h1('AWS CLI Topic Guide')

    def doc_description(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Description')
        doc.include_doc_string(self.DESCRIPTION)
        doc.style.new_paragraph()

    def doc_synopsis_start(self, help_command, **kwargs):
        pass

    def doc_synopsis_end(self, help_command, **kwargs):
        pass

    def doc_options_start(self, help_command, **kwargs):
        pass

    def doc_options_end(self, help_command, **kwargs):
        pass

    def doc_global_option(self, help_command, **kwargs):
        pass

    def doc_subitems_start(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.h2('Available Topics')

        categories = self._topic_tag_db.query('category')
        topic_names = self._topic_tag_db.get_all_topic_names()

        # Sort the categories
        category_names = sorted(categories.keys())
        for category_name in category_names:
            doc.style.h3(category_name)
            doc.style.new_paragraph()
            # Write out the topic and a description for each topic under
            # each category.
            for topic_name in sorted(categories[category_name]):
                description = self._topic_tag_db.get_tag_single_value(
                    topic_name, 'description'
                )
                doc.write('* ')
                doc.style.sphinx_reference_label(
                    label=f'cli:aws help {topic_name}', text=topic_name
                )
                doc.write(f': {description}\n')
        # Add a hidden toctree to make sure everything is connected in
        # the document.
        doc.style.hidden_toctree()
        for topic_name in topic_names:
            doc.style.hidden_tocitem(topic_name)


class TopicDocumentEventHandler(TopicListerDocumentEventHandler):
    def doc_breadcrumbs(self, help_command, **kwargs):
        doc = help_command.doc
        if doc.target != 'man':
            doc.write('[ ')
            doc.style.sphinx_reference_label(label='cli:aws', text='aws')
            doc.write(' . ')
            doc.style.sphinx_reference_label(
                label='cli:aws help topics', text='topics'
            )
            doc.write(' ]')

    def doc_title(self, help_command, **kwargs):
        doc = help_command.doc
        doc.style.new_paragraph()
        doc.style.link_target_definition(
            refname=f'cli:aws help {self.help_command.name}', link=''
        )
        title = self._topic_tag_db.get_tag_single_value(
            help_command.name, 'title'
        )
        doc.style.h1(title)

    def doc_description(self, help_command, **kwargs):
        doc = help_command.doc
        topic_filename = os.path.join(
            self._topic_tag_db.topic_dir, help_command.name + '.rst'
        )
        contents = self._remove_tags_from_content(topic_filename)
        doc.writeln(contents)
        doc.style.new_paragraph()

    def _remove_tags_from_content(self, filename):
        with open(filename) as f:
            lines = f.readlines()

        content_begin_index = 0
        for i, line in enumerate(lines):
            # If a line is encountered that does not begin with the tag
            # end the search for tags and mark where tags end.
            if not self._line_has_tag(line):
                content_begin_index = i
                break

        # Join all of the non-tagged lines back together.
        return ''.join(lines[content_begin_index:])

    def _line_has_tag(self, line):
        for tag in self._topic_tag_db.valid_tags:
            if line.startswith(':' + tag + ':'):
                return True
        return False

    def doc_subitems_start(self, help_command, **kwargs):
        pass


class GlobalOptionsDocumenter:
    """Documenter used to pre-generate global options docs."""

    def __init__(self, help_command):
        self._help_command = help_command

    def _remove_multilines(self, s):
        return re.sub(r'\n+', '\n', s)

    def doc_global_options(self):
        help_command = self._help_command
        for arg in help_command.arg_table:
            argument = help_command.arg_table.get(arg)
            help_command.doc.writeln(
                f"``{argument.cli_name}`` ({argument.cli_type_name})"
            )
            help_command.doc.style.indent()
            help_command.doc.style.new_paragraph()
            help_command.doc.include_doc_string(argument.documentation)
            if argument.choices:
                help_command.doc.style.start_ul()
                for choice in argument.choices:
                    help_command.doc.style.li(choice)
                help_command.doc.style.end_ul()
            help_command.doc.style.dedent()
            help_command.doc.style.new_paragraph()
        global_options = help_command.doc.getvalue().decode('utf-8')
        return self._remove_multilines(global_options)

    def doc_global_synopsis(self):
        help_command = self._help_command
        for arg in help_command.arg_table:
            argument = help_command.arg_table.get(arg)
            if argument.cli_type_name == 'boolean':
                arg_synopsis = f"[{argument.cli_name}]"
            else:
                arg_synopsis = f"[{argument.cli_name} <value>]"
            help_command.doc.writeln(arg_synopsis)
        global_synopsis = help_command.doc.getvalue().decode('utf-8')
        return self._remove_multilines(global_synopsis)



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand
from awscli.customizations.utils import uni_print


def register_dsql_customizations(cli):
    cli.register(
        'building-command-table.dsql', _add_generate_dsql_db_connect_auth_token
    )
    cli.register(
        'building-command-table.dsql',
        _add_generate_dsql_db_connect_admin_auth_token,
    )


def _add_generate_dsql_db_connect_auth_token(command_table, session, **kwargs):
    command = GenerateDBConnectAuthTokenCommand(session)
    command_table['generate-db-connect-auth-token'] = command


def _add_generate_dsql_db_connect_admin_auth_token(
    command_table, session, **kwargs
):
    command = GenerateDBConnectAdminAuthTokenCommand(session)
    command_table['generate-db-connect-admin-auth-token'] = command


class GenerateDBConnectAuthTokenCommand(BasicCommand):
    NAME = 'generate-db-connect-auth-token'
    DESCRIPTION = 'Generates an authorization token used to connect to a DSQL database with IAM credentials.'
    ARG_TABLE = [
        {
            'name': 'hostname',
            'required': True,
            'help_text': 'Cluster endpoint e.g. http://test.example.com',
        },
        {
            'name': 'expires-in',
            'cli_type_name': 'integer',
            'default': 900,
            'required': False,
            'help_text': 'Token expiry duration in seconds e.g. 3600. Default is 900 seconds.',
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        dsql = self._session.create_client(
            'dsql',
            parsed_globals.region,
            parsed_globals.endpoint_url,
            parsed_globals.verify_ssl,
        )

        token = dsql.generate_db_connect_auth_token(
            parsed_args.hostname, parsed_globals.region, parsed_args.expires_in
        )
        uni_print(token)
        uni_print('\n')
        return 0


class GenerateDBConnectAdminAuthTokenCommand(BasicCommand):
    NAME = 'generate-db-connect-admin-auth-token'
    DESCRIPTION = 'Generates an Admin authorization token used to connect to a DSQL database with IAM credentials.'
    ARG_TABLE = [
        {
            'name': 'hostname',
            'required': True,
            'help_text': 'Cluster endpoint e.g. http://test.example.com',
        },
        {
            'name': 'expires-in',
            'cli_type_name': 'integer',
            'default': 900,
            'required': False,
            'help_text': 'Token expiry duration in seconds e.g. 3600. Default is 900 seconds.',
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        dsql = self._session.create_client(
            'dsql',
            parsed_globals.region,
            parsed_globals.endpoint_url,
            parsed_globals.verify_ssl,
        )

        token = dsql.generate_db_connect_admin_auth_token(
            parsed_args.hostname, parsed_globals.region, parsed_args.expires_in
        )
        uni_print(token)
        uni_print('\n')
        return 0



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class CliDriverFetcher:
    def __init__(self, cli_driver):
        self._cli_driver = cli_driver

    def _get_command(self, lineage, current_command):
        subcommand_table = self._cli_driver.subcommand_table
        for arg in lineage[1:]:
            subcommand = subcommand_table[arg]
            subcommand_table = subcommand.subcommand_table
        return subcommand_table.get(current_command)

    def _get_argument(self, lineage, current_command, arg_name):
        command = self._get_command(lineage, current_command)
        if command:
            return command.arg_table.get(arg_name)

    def get_operation_model(self, lineage, current_command):
        if len(lineage) > 1 and current_command:
            command = self._get_command(lineage, current_command)
            return command.create_help_command().obj

    def get_argument_model(self, lineage, current_command, arg_name):
        return getattr(
            self._get_argument(lineage, current_command, arg_name),
            'argument_model',
            None,
        )

    def get_argument_documentation(self, lineage, current_command, arg_name):
        return getattr(
            self._get_argument(lineage, current_command, arg_name),
            'documentation',
            '',
        )

    def get_global_arg_documentation(self, arg_name):
        return self._cli_driver.arg_table[arg_name].documentation

    def get_global_arg_choices(self, arg_name):
        if arg_name in self._cli_driver.arg_table:
            return self._cli_driver.arg_table[arg_name].choices



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

from awscli.customizations.commands import BasicCommand


def register_get_status(cli):
    cli.register('building-command-table.configservice', add_get_status)


def add_get_status(command_table, session, **kwargs):
    command_table['get-status'] = GetStatusCommand(session)


class GetStatusCommand(BasicCommand):
    NAME = 'get-status'
    DESCRIPTION = (
        'Reports the status of all of configuration '
        'recorders and delivery channels.'
    )

    def __init__(self, session):
        self._config_client = None
        super(GetStatusCommand, self).__init__(session)

    def _run_main(self, parsed_args, parsed_globals):
        self._setup_client(parsed_globals)
        self._check_configuration_recorders()
        self._check_delivery_channels()
        return 0

    def _setup_client(self, parsed_globals):
        client_args = {
            'verify': parsed_globals.verify_ssl,
            'region_name': parsed_globals.region,
            'endpoint_url': parsed_globals.endpoint_url,
        }
        self._config_client = self._session.create_client(
            'config', **client_args
        )

    def _check_configuration_recorders(self):
        status = self._config_client.describe_configuration_recorder_status()
        sys.stdout.write('Configuration Recorders:\n\n')
        for configuration_recorder in status['ConfigurationRecordersStatus']:
            self._check_configure_recorder_status(configuration_recorder)
            sys.stdout.write('\n')

    def _check_configure_recorder_status(self, configuration_recorder):
        # Get the name of the recorder and print it out.
        name = configuration_recorder['name']
        sys.stdout.write('name: %s\n' % name)

        # Get the recording status and print it out.
        recording = configuration_recorder['recording']
        recording_map = {False: 'OFF', True: 'ON'}
        sys.stdout.write('recorder: %s\n' % recording_map[recording])

        # If the recorder is on, get the last status and print it out.
        if recording:
            self._check_last_status(configuration_recorder)

    def _check_delivery_channels(self):
        status = self._config_client.describe_delivery_channel_status()
        sys.stdout.write('Delivery Channels:\n\n')
        for delivery_channel in status['DeliveryChannelsStatus']:
            self._check_delivery_channel_status(delivery_channel)
            sys.stdout.write('\n')

    def _check_delivery_channel_status(self, delivery_channel):
        # Get the name of the delivery channel and print it out.
        name = delivery_channel['name']
        sys.stdout.write('name: %s\n' % name)

        # Obtain the various delivery statuses.
        stream_delivery = delivery_channel['configStreamDeliveryInfo']
        history_delivery = delivery_channel['configHistoryDeliveryInfo']
        snapshot_delivery = delivery_channel['configSnapshotDeliveryInfo']

        # Print the statuses out if they exist.
        if stream_delivery:
            self._check_last_status(stream_delivery, 'stream delivery ')
        if history_delivery:
            self._check_last_status(history_delivery, 'history delivery ')
        if snapshot_delivery:
            self._check_last_status(snapshot_delivery, 'snapshot delivery ')

    def _check_last_status(self, status, status_name=''):
        last_status = status['lastStatus']
        sys.stdout.write('last %sstatus: %s\n' % (status_name, last_status))
        if last_status == "FAILURE":
            sys.stdout.write('error code: %s\n' % status['lastErrorCode'])
            sys.stdout.write('message: %s\n' % status['lastErrorMessage'])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


# This is the main entry point for auto-completion.  This is imported
# everytime a user hits <TAB>.  Try to avoid any expensive module level
# work or really heavyweight imports.  Prefer to lazy load as much as possible.

from awscli.autocomplete import completer, custom, filters, parser, serverside
from awscli.autocomplete.local import basic, fetcher, model


def create_autocompleter(
    index_filename=None,
    custom_completers=None,
    driver=None,
    response_filter=None,
):
    if response_filter is None:
        response_filter = filters.startswith_filter
    if custom_completers is None:
        custom_completers = custom.get_custom_completers()
    index = model.ModelIndex(index_filename)
    cli_parser = parser.CLIParser(index)
    cli_driver_fetcher = None
    if driver is not None:
        cli_driver_fetcher = fetcher.CliDriverFetcher(driver)
    completers = [
        basic.RegionCompleter(response_filter=response_filter),
        basic.ProfileCompleter(response_filter=response_filter),
        basic.ModelIndexCompleter(
            index, cli_driver_fetcher, response_filter=response_filter
        ),
        basic.FilePathCompleter(response_filter=response_filter),
        serverside.create_server_side_completer(
            index_filename, response_filter=response_filter
        ),
        basic.ShorthandCompleter(
            cli_driver_fetcher, response_filter=response_filter
        ),
        basic.QueryCompleter(
            cli_driver_fetcher, response_filter=response_filter
        ),
    ] + custom_completers
    cli_completer = completer.AutoCompleter(cli_parser, completers)
    return cli_completer


def autocomplete(command_line, position=None):
    completer = create_autocompleter()
    results = completer.autocomplete(command_line, position)
    print("\n".join([result.name for result in results]))



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import sys

from awscli.customizations.commands import BasicCommand
from awscli.customizations.s3.utils import find_bucket_key
from awscli.customizations.utils import s3_bucket_exists

S3_BUCKET = {
    'name': 's3-bucket',
    'required': True,
    'help_text': (
        'The S3 bucket that the AWS Config delivery channel'
        ' will use. If the bucket does not exist, it will '
        'be automatically created. The value for this '
        'argument should follow the form '
        'bucket/prefix. Note that the prefix is optional.'
    ),
}

SNS_TOPIC = {
    'name': 'sns-topic',
    'required': True,
    'help_text': (
        'The SNS topic that the AWS Config delivery channel'
        ' will use. If the SNS topic does not exist, it '
        'will be automatically created. Value for this '
        'should be a valid SNS topic name or the ARN of an '
        'existing SNS topic.'
    ),
}

IAM_ROLE = {
    'name': 'iam-role',
    'required': True,
    'help_text': (
        'The IAM role that the AWS Config configuration '
        'recorder will use to record current resource '
        'configurations. Value for this should be the '
        'ARN of the desired IAM role.'
    ),
}


def register_subscribe(cli):
    cli.register('building-command-table.configservice', add_subscribe)


def add_subscribe(command_table, session, **kwargs):
    command_table['subscribe'] = SubscribeCommand(session)


class SubscribeCommand(BasicCommand):
    NAME = 'subscribe'
    DESCRIPTION = (
        'Subscribes user to AWS Config by creating an AWS Config '
        'delivery channel and configuration recorder to track '
        'AWS resource configurations. The names of the default '
        'channel and configuration recorder will be default.'
    )
    ARG_TABLE = [S3_BUCKET, SNS_TOPIC, IAM_ROLE]

    def __init__(self, session):
        self._s3_client = None
        self._sns_client = None
        self._config_client = None
        super(SubscribeCommand, self).__init__(session)

    def _run_main(self, parsed_args, parsed_globals):
        # Setup the necessary all of the necessary clients.
        self._setup_clients(parsed_globals)

        # Prepare a s3 bucket for use.
        s3_bucket_helper = S3BucketHelper(self._s3_client)
        bucket, prefix = s3_bucket_helper.prepare_bucket(parsed_args.s3_bucket)

        # Prepare a sns topic for use.
        sns_topic_helper = SNSTopicHelper(self._sns_client)
        sns_topic_arn = sns_topic_helper.prepare_topic(parsed_args.sns_topic)

        name = 'default'

        # Create a configuration recorder.
        self._config_client.put_configuration_recorder(
            ConfigurationRecorder={
                'name': name,
                'roleARN': parsed_args.iam_role,
            }
        )

        # Create a delivery channel.
        delivery_channel = {
            'name': name,
            's3BucketName': bucket,
            'snsTopicARN': sns_topic_arn,
        }

        if prefix:
            delivery_channel['s3KeyPrefix'] = prefix

        self._config_client.put_delivery_channel(
            DeliveryChannel=delivery_channel
        )

        # Start the configuration recorder.
        self._config_client.start_configuration_recorder(
            ConfigurationRecorderName=name
        )

        # Describe the configuration recorders
        sys.stdout.write('Subscribe succeeded:\n\n')
        sys.stdout.write('Configuration Recorders: ')
        response = self._config_client.describe_configuration_recorders()
        sys.stdout.write(
            json.dumps(response['ConfigurationRecorders'], indent=4)
        )
        sys.stdout.write('\n\n')

        # Describe the delivery channels
        sys.stdout.write('Delivery Channels: ')
        response = self._config_client.describe_delivery_channels()
        sys.stdout.write(json.dumps(response['DeliveryChannels'], indent=4))
        sys.stdout.write('\n')

        return 0

    def _setup_clients(self, parsed_globals):
        client_args = {
            'verify': parsed_globals.verify_ssl,
            'region_name': parsed_globals.region,
        }
        self._s3_client = self._session.create_client('s3', **client_args)
        self._sns_client = self._session.create_client('sns', **client_args)
        # Use the specified endpoint only for config related commands.
        client_args['endpoint_url'] = parsed_globals.endpoint_url
        self._config_client = self._session.create_client(
            'config', **client_args
        )


class S3BucketHelper:
    def __init__(self, s3_client):
        self._s3_client = s3_client

    def prepare_bucket(self, s3_path):
        bucket, key = find_bucket_key(s3_path)
        bucket_exists = self._check_bucket_exists(bucket)
        if not bucket_exists:
            self._create_bucket(bucket)
            sys.stdout.write('Using new S3 bucket: %s\n' % bucket)
        else:
            sys.stdout.write('Using existing S3 bucket: %s\n' % bucket)
        return bucket, key

    def _check_bucket_exists(self, bucket):
        return s3_bucket_exists(self._s3_client, bucket)

    def _create_bucket(self, bucket):
        region_name = self._s3_client.meta.region_name
        params = {'Bucket': bucket}
        bucket_config = {'LocationConstraint': region_name}
        if region_name != 'us-east-1':
            params['CreateBucketConfiguration'] = bucket_config
        self._s3_client.create_bucket(**params)


class SNSTopicHelper:
    def __init__(self, sns_client):
        self._sns_client = sns_client

    def prepare_topic(self, sns_topic):
        sns_topic_arn = sns_topic
        # Create the topic if a name is given.
        if not self._check_is_arn(sns_topic):
            response = self._sns_client.create_topic(Name=sns_topic)
            sns_topic_arn = response['TopicArn']
            sys.stdout.write('Using new SNS topic: %s\n' % sns_topic_arn)
        else:
            sys.stdout.write('Using existing SNS topic: %s\n' % sns_topic_arn)
        return sns_topic_arn

    def _check_is_arn(self, sns_topic):
        # The name of topic cannot contain a colon only arns have colons.
        return ':' in sns_topic
